#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tensorflow Graph Utils Helper Classes."""

import copy
import logging
import re
from collections import namedtuple

import numpy as np
from tensorflow.core.framework import attr_value_pb2, graph_pb2, node_def_pb2
from tensorflow.python.framework import tensor_util

from neural_compressor.tensorflow.utils import singleton

logger = logging.getLogger("neural_compressor")


@singleton
class GraphAnalyzer:
    """Tensorflow Graph Analyzer class which implemented under singleton mode.

    This class provides the following API:
    * Analyze the graph
    * Analyze the input/output node names of the specified graph
    """

    # TODO add the positive input flag
    node_details = namedtuple("node_details", ["node", "outputs"])

    def __init__(self, extend_engine=None):
        """Initialization.

        Args:
            extend_engine: extended engine, for future extension API。
        """
        self._graph = None
        self.extend_engine = extend_engine

    @property
    def graph(self):
        """Getter of the _graph object.

        Returns:
            graph: current graphdef object
        """
        return self._graph

    @graph.setter
    def graph(self, new_graph):
        """Update the internal graph value.

        Args:
            new_graph (graphdef object): new model object
        """
        self._graph = new_graph

    def _has_positive_input(self, start_node):
        """Check the start_node if has positive input."""
        op_type = start_node.op
        if op_type in ("Relu", "Relu6") or op_type.find("AndRelu") != -1:
            return True
        elif op_type.startswith("Quantized") and not op_type.endswith("AndRelu"):
            return False
        elif op_type in ("Concat", "Add", "AddV2", "AddN"):
            for each_input in start_node.input:
                has_relu = self._has_positive_input(
                    self.node_name_details[GraphRewriterHelper.node_name_from_input(each_input)].node
                )
                if not has_relu:
                    return False
            return True
        elif op_type in (
            "Conv3D",
            "Conv2D",
            "DepthwiseConv2D",
            "QuantizeV2",
            "DepthwiseConv2dNative",
            "MaxPool",
            "MaxPool3D",
            "Requantize",
            "AvgPool",
            "Pad",
            "CropAndResize",
            "Dequantize",
            "Mean",
            "MatMul",
            "FusedBatchNormV3",
            "_MklFusedInstanceNorm",
        ):
            return self._has_positive_input(
                self.node_name_details[GraphRewriterHelper.node_name_from_input(start_node.input[0])].node
            )
        else:
            return False

    def has_positive_input(self, node_name):
        """Check the specified node has positive input data or not.

        Args:
            node_name (string): node name

        Returns:
            bool: return True if the node has the positive input data,
                return False if the node has the negative input data.
        """
        return self._has_positive_input(self.node_name_details[node_name].node)

    def get_graph_input_output(self):
        """Get the graphdef input/output node names.

        Sometimes, the configuration doesn't specifies the input/output names of the graph,
        but tensorflow need to know them clearly to run the graph.We implement this function has the similar
        feature like summarize_graph.py which writtern by Google.

        Returns:
            tuple: (inputs' name list, outputs'name list)
        """
        input_node_names = []
        output_node_names = []
        unlikely_output_types = [
            "Const",
            "HostConst",
            "Assign",
            "NoOp",
            "Parameter",
            "Assert",
            "save",
            "global_step",
            "read",
            "switch",
            "cond",
            "train",
            "init_ops",
            "[A-Za-z]+Dataset",
        ]
        unlikely_input_types = [
            "FIFOQueueV2",
            "QueueDequeueV2",
            "QueueDequeueUpToV2",
            "OneShotIterator",
            "IteratorGetNext",
            "IteratorV2",
        ]
        exclude_input_names = []
        extra_input_names = []

        for _, i in self.node_name_details.items():
            for exclude_input_name in exclude_input_names:
                if exclude_input_name == i.node.name:
                    if i.node.op in unlikely_input_types:
                        exclude_input_names += i.outputs
                    else:
                        extra_input_names.append(i.node.name)
            if i.node.op in ["Const", "HostConst", "Variable", "VariableV2"]:
                continue
            if not i.node.input and not i.outputs:
                logger.debug("Skip isolated node {}.".format(i.node.name))
            elif i.node.op == "Placeholder":
                input_node_names.append(i.node.name)
            elif not i.node.input:
                if i.node.op not in unlikely_input_types:
                    input_node_names.append(i.node.name)
                else:
                    exclude_input_names += i.outputs
            elif (
                not i.outputs
                and i.node.op not in unlikely_output_types
                and not re.match(unlikely_output_types[-1], i.node.op)
            ):
                output_node_names.append(i.node.name)
            else:
                pass

        if len(input_node_names) == 0 and len(extra_input_names) != 0:
            for extra_input_name in extra_input_names:
                input_node_names.append(extra_input_name)

        logger.warning(
            "Found possible input node names: {}, output node names: {}.".format(input_node_names, output_node_names)
        )

        return (input_node_names, output_node_names)

    def query_fusion_pattern_nodes(self, patterns=None):
        """Public interface for query the nodes aggregation status.

        Args:
            patterns (string list): Please check the _search_patterns definition.

        Returns:
            [string list]: The matched node names which saved as the string list.
        """
        if self.extend_engine:
            # Todo keep this for future extension API
            pass
        else:
            return self._search_patterns(patterns)

    def _search_patterns(self, input_pattern):
        """Search user specified patterns on internal grpah structure.

        Args:
            input_pattern (list): The element of the pattern list could be string/list/tuple.
            string or list means the specified types are mandatory while tuple stands for optional.
            e.g:
            ['Conv2D', ['BiasAdd'], ("Add", "AddN"), ["Relu","Relu6"]] it equals to below patterns:
            Conv2D + BiasAdd + Add + Relu
            Conv2D + BiasAdd + AddN + Relu
            Conv2D + BiasAdd + Add + Relu6
            Conv2D + BiasAdd + AddN + Relu6
            Conv2D + BiasAdd + Relu
            Conv2D + BiasAdd + Relu6

        Return: [string list]. Each matched pattern composed of matched node name and we put the
                    match node op as the last element of each pair.
                    e.g
                    [
                        ['resnet_model/conv2d_4/Conv2D',
                        'resnet_model/batch_normalization_4/FusedBatchNorm',
                        'resnet_model/add',
                        'resnet_model/Relu_3',
                        ['Conv2D', 'BiasAdd', 'Add', 'Relu']],
                        ['resnet_model/conv2d_7/Conv2D',
                        'resnet_model/batch_normalization_7/FusedBatchNorm',
                        'resnet_model/add_1',
                        'resnet_model/Relu_6',
                        ['Conv2D', 'BiasAdd', 'AddN', 'Relu6']]
                    ]
        """

        def _validate_input(data, criteria):
            if isinstance(criteria, str) and data == criteria:
                return True

            if isinstance(criteria, (list, tuple)) and data in criteria:
                return True

            return False

        def _compare_list(list_a, list_b):
            """Check list a is a subset of list b.

            e.g, list a is ['a', 'b', 'c'] while list b is ['a', 'b', 'c', 'd'],
            then list a is subset of list b.

            Args:
                list_a ([Any]): list A
                list_b ([Any]): list B

            Returns:
                [bool]: list a is a subset of list b or not.
            """
            assert isinstance(list_a, list)
            assert isinstance(list_b, list)
            is_subset = True

            for index, value in enumerate(list_a):
                is_subset &= value == list_b[index]

            return is_subset

        def _dfs(op_names, op_types, graph_info, node, pattern):
            if pattern == []:
                return
            start_index = 0
            end_index = len(pattern) - 1
            matched_flag = False
            while start_index <= end_index:
                matched_flag = _validate_input(node.op, pattern[end_index])

                if not matched_flag and isinstance(pattern[end_index], tuple):
                    end_index -= 1
                    continue

                if matched_flag:
                    op_names.append(node.name)
                    op_types.append(node.op)
                    break

                return

            if start_index == end_index:
                if matched_flag:
                    matched_res = copy.deepcopy(op_names)
                    matched_res.reverse()
                    op_types_copy = copy.deepcopy(op_types)
                    op_types_copy.reverse()
                    matched_res.append(op_types_copy)
                    if matched_res not in output_result:
                        output_result.append(matched_res)

                    op_names.pop()
                    op_types.pop()
                return

            for index, value in enumerate(node.input):
                cur_node = graph_info[GraphRewriterHelper.node_name_from_input(value)].node
                _dfs(op_names, op_types, graph_info, cur_node, pattern[:end_index])
                if index == len(node.input) - 1:
                    op_names.pop()
                    op_types.pop()

        output_result = []

        for _, v in self.node_name_details.items():
            start_index = len(input_pattern) - 1
            while start_index >= 0:
                find_first_match = _validate_input(v.node.op, input_pattern[start_index])
                if find_first_match:
                    break

                if isinstance(input_pattern[start_index], tuple):
                    start_index -= 1
                    continue

                start_index = -2

            if start_index < 0:
                continue

            visited_op_name = []
            visited_op_types = []

            _dfs(visited_op_name, visited_op_types, self.node_name_details, v.node, input_pattern)

        sorted_output = sorted(output_result, key=lambda i: i[-1])

        useless_match_list = []
        for index, value in enumerate(sorted_output):
            if index == len(sorted_output) - 1:
                break

            next_matched_op_names = sorted_output[index + 1][:-1]
            if len(value[:-1]) < len(next_matched_op_names) and _compare_list(value[:-1], next_matched_op_names):
                useless_match_list.append(value)

        for i in useless_match_list:
            sorted_output.remove(i)

        longest_match = {}
        final_output = []
        for i in sorted_output:
            key = i[0]
            if key not in longest_match:
                longest_match[key] = i[-1]
                continue

            if len(longest_match[key]) < len(i[-1]):
                longest_match[key] = i[-1]

        for i in sorted_output:
            if i[0] in longest_match and i[-1] == longest_match[i[0]]:
                final_output.append(i)

        return final_output

    def remove_node_with_single_input_output(self, node_name):
        """Remove node with one input and rebuild internal graph data structure.

        Args:
            node_name (string): node name

        Returns:
            [bool]: True if remove the node without exception,
                    False if failed to remove it.
        """
        if node_name not in self.node_name_details:
            logger.debug("The {} is not a valid node name.".format(node_name))
            return False

        non_const_node_count = len(
            [
                GraphRewriterHelper.node_name_from_input(i)
                for i in self.node_name_details[node_name].node.input
                if self.node_name_details[GraphRewriterHelper.node_name_from_input(i)].node.op != "Const"
            ]
        )

        if non_const_node_count > 1:
            logger.debug("The target node {} has more than one input.".format(node_name))
            return False

        try:
            top_node_name = GraphRewriterHelper.node_name_from_input(self.node_name_details[node_name].node.input[0])

            for bottom_node_name in self.node_name_details[node_name].outputs:
                update_output_name = [
                    bottom_node_name if i == node_name else i for i in self.node_name_details[top_node_name].outputs
                ]
                self.node_name_details[top_node_name]._replace(outputs=update_output_name)

                update_input_name = [
                    self.node_name_details[node_name].node.input[0] if i == node_name else i
                    for i in self.node_name_details[bottom_node_name].node.input
                ]

                if self.node_name_details[bottom_node_name].node.input:
                    self.node_name_details[bottom_node_name].node.ClearField("input")
                    self.node_name_details[bottom_node_name].node.input.extend(update_input_name)

        except Exception as e:
            logger.debug("Fail to remove node {} due to {}.".format(node_name, str(e)))
            return False
        else:
            return self.remove_node(node_name)

    def remove_node(self, node_name):
        """Remove the user specified node by its name.

        Args:
            node_name (string): node name string.

        Returns:
            [bool]: True if remove the node without exception.
                    False if failed to remove it.
        """
        if node_name not in self.node_name_details:
            logger.debug("The {} is not a valid node name.".format(node_name))
            return False
        try:
            self.node_name_details.pop(node_name)
        except Exception as e:
            logger.info("Fail to remove {} due to {}.".format(node_name, str(e)))
            return False
        else:
            logger.debug("{} has been removed.".format(node_name))
            return True

    def replace_const_node(self, new_const_node, target_node, old_constant_node_name, replace_all=True):
        """Replace the specified const node with another one.

        Args:
            new_const_node (NodeDef): node name string.
            target_node (list): the string list that contains name of node that
                                need to be replaced const node.
            old_constant_node_name (string): the outdated const node name.
            replace_all (bool): replace the specified node name once or not.
        """
        new_const_node_name = new_const_node.name

        self.node_name_details[new_const_node_name] = self.node_details(node=new_const_node, outputs=target_node)

        for sub_node in target_node:
            if sub_node not in self.node_name_details:
                continue
            for index, each_node_name in enumerate(self.node_name_details[sub_node].node.input):
                if each_node_name + ":0" == old_constant_node_name or each_node_name == old_constant_node_name:
                    new_input_name = (
                        self.node_name_details[sub_node].node.input[:index]
                        + [new_const_node_name]
                        + self.node_name_details[sub_node].node.input[index + 1 :]
                    )
                    self.node_name_details[sub_node].node.ClearField("input")
                    self.node_name_details[sub_node].node.input.extend(new_input_name)
                    if old_constant_node_name in self.node_name_details:
                        self.node_name_details[old_constant_node_name].outputs.remove(sub_node)
                        if len(self.node_name_details[old_constant_node_name].outputs) == 0:
                            self.remove_node(old_constant_node_name)
                    if not replace_all:
                        break

    def replace_constant_graph_with_constant_node(self, new_node, old_end_node_name):
        """Remove sub-graph with a const node.

        Args:
            new_node (nodedef): the constant node
            old_end_node_name (string):  the sub-graph end node which will be updated by new node

        Returns:
            [bool]: True if remove the node without exception.
                    False if failed to remove it.
        """
        new_node_name = new_node.name

        if new_node.op != "Const":
            logger.warning("The input of replace_with_constant_node must be a constant node.")
            return False
        try:
            inputs = self.node_name_details[old_end_node_name].node.input
            inputs = [GraphRewriterHelper.node_name_from_input(i) for i in inputs]
            for input_name in inputs:
                if self.node_name_details[input_name].node.op != "Const":
                    logger.warning("The subgraph replaces must be constant.")
                    return False
                elif len(self.node_name_details[input_name].outputs) == 1:
                    self.node_name_details.pop(input_name)
            output_node_name = self.node_name_details[old_end_node_name].outputs
            self.replace_node(new_node, old_end_node_name, output_node_name)
            self.node_name_details[new_node_name].node.ClearField("input")
        except Exception as e:
            logger.info("Fail to replace {} due to {}.".format(old_end_node_name, str(e)))
            return False
        else:
            return True

    def replace_single_node(
        self, new_node, old_output_node_names, old_output_name, old_input_node_names, old_input_name
    ):
        """Insert one node into the graph.

        Args:
            new_node (nodedef): new nodedef object
            old_output_node_names (string list):the node names that would be the top node of new
                                                node.
            old_output_name (string list): the names that need to be updated with new node name
            old_input_node_names (string list): the node names that would be the bottom node of new
                                                node.
            old_input_name (string list): the names that need to be updated with new node name
        """
        new_node_name = new_node.name
        for i in old_output_node_names:
            while old_output_name in self.node_name_details[i].outputs:
                self.node_name_details[i].outputs.remove(old_output_name)
            self.node_name_details[i].outputs.append(new_node_name)

        self.node_name_details[new_node_name] = self.node_details(node=new_node, outputs=old_input_node_names)

        for each_input_node_name in old_input_node_names:
            for index, each_node_name in enumerate(self.node_name_details[each_input_node_name].node.input):
                if self.node_name_details[each_input_node_name].node.input and (each_node_name) == old_input_name:
                    new_input_name = (
                        self.node_name_details[each_input_node_name].node.input[:index]
                        + [new_node_name]
                        + self.node_name_details[each_input_node_name].node.input[index + 1 :]
                    )
                    self.node_name_details[each_input_node_name].node.ClearField("input")
                    self.node_name_details[each_input_node_name].node.input.extend(new_input_name)

    def replace_node(self, new_node, old_node_name, output_nodes_name):
        """Replace the node into the internal data structure node_name_details.

        Args:
            new_node (nodedef): the nodedef object.
            old_node_name (string): the parent node of input node.
            output_nodes_name (string list): output node names list
        """
        new_node_name = new_node.name
        self.node_name_details[new_node_name] = self.node_details(node=new_node, outputs=output_nodes_name)
        old_node = self.node_name_details[old_node_name].node
        for input_node_name in old_node.input:
            if input_node_name in self.node_name_details:
                self.node_name_details[input_node_name].outputs.remove(old_node_name)
                self.node_name_details[input_node_name].outputs.append(new_node_name)

        for node_name in output_nodes_name:
            for index, each_node_name in enumerate(self.node_name_details[node_name].node.input):
                if (
                    self.node_name_details[node_name].node.input
                    and GraphRewriterHelper.node_name_from_input(each_node_name) == old_node_name
                ):
                    new_input_name = (
                        self.node_name_details[node_name].node.input[:index]
                        + [new_node_name]
                        + self.node_name_details[node_name].node.input[index + 1 :]
                    )
                    self.node_name_details[node_name].node.ClearField("input")
                    self.node_name_details[node_name].node.input.extend(new_input_name)
        self.remove_node(old_node_name)

    def add_node(self, new_node, start_node_name, end_node_names):
        """Add the node into the internal data structure node_name_details.

        Args:
            new_node (nodedef): the nodedef object.
            start_node_name (string): the parent node of input node.
            end_node_names (string list): output node names list
        """
        new_node_name = new_node.name

        if new_node_name in self.node_name_details:
            logger.debug("Remove the existed node {} from internal data structure.".format((new_node_name)))
            self.node_name_details.pop(new_node_name)

        self.node_name_details[new_node_name] = self.node_details(node=new_node, outputs=end_node_names)

        for end_node_name in end_node_names:
            # Update start node's output info
            if end_node_name not in self.node_name_details:
                continue
            if (
                start_node_name
                and end_node_name
                in self.node_name_details[GraphRewriterHelper.node_name_from_input(start_node_name)].outputs
            ):
                self.node_name_details[GraphRewriterHelper.node_name_from_input(start_node_name)].outputs.remove(
                    end_node_name
                )

            # reset output node's input
            for index, each_node_name in enumerate(self.node_name_details[end_node_name].node.input):
                if each_node_name == start_node_name:
                    new_input_name = (
                        self.node_name_details[end_node_name].node.input[:index]
                        + [new_node_name]
                        + self.node_name_details[end_node_name].node.input[index + 1 :]
                    )
                    self.node_name_details[end_node_name].node.ClearField("input")
                    self.node_name_details[end_node_name].node.input.extend(new_input_name)

        # add the inserted node into the start node's output.
        if start_node_name:
            self.node_name_details[GraphRewriterHelper.node_name_from_input(start_node_name)].outputs.append(
                new_node_name
            )

    def dump_graph(self):
        """Dump the current model's graphdef.

        Returns:
            [graphdef]: A graphdef object
        """
        output_graph_def = graph_pb2.GraphDef()
        for _, v in self.node_name_details.items():
            output_graph_def.node.extend([v.node])

        return output_graph_def

    def get_frame_info(self):
        """Get the frame info of the model.

        Returns:
            [parent_frame_details]: OrderedDict frame info of the graph nodes.
        """
        from collections import OrderedDict

        self.parent_frame_details = OrderedDict()
        input_node_names, _ = self.get_graph_input_output()

        traverse_list = copy.deepcopy(input_node_names)
        visited = []

        while traverse_list:
            node_name = traverse_list.pop(0)
            node_details = self.node_name_details[node_name]

            if node_details.node.name in visited:
                continue

            for output in node_details.outputs:
                traverse_list.append(output)

                inputs = node_details.node.input
                if not inputs:
                    self.parent_frame_details[node_details.node.name] = None
                if self.node_name_details[output].node.op == "Enter":
                    self.parent_frame_details[output] = self.node_name_details[output].node
                elif self.node_name_details[output].node.op == "Exit":
                    self.parent_frame_details[output] = None
                else:
                    if output in self.parent_frame_details and self.parent_frame_details[output]:
                        if (
                            node_details.node.name in self.parent_frame_details
                            and self.parent_frame_details[node_details.node.name]
                        ):
                            assert (
                                self.parent_frame_details[output].attr["frame_name"]
                                == self.parent_frame_details[node_details.node.name].attr["frame_name"]
                            )
                    else:
                        if node_details.node.name in self.parent_frame_details:
                            self.parent_frame_details[output] = self.parent_frame_details[node_details.node.name]

            visited.append(node_details.node.name)
        return self.parent_frame_details

    def parse_graph(self, input_graph_def=None):
        """Analyze the input graphdef and return the list contains each node's input/outputnode names.

        Args:
            input_graph_def ([graphdef]): graphdef object

        Returns:
            [list]: A list contains each node's inputs/outputs info.
        """
        if not input_graph_def:
            input_graph_def = self._graph

        self.node_name_details = {}

        for node in input_graph_def.node:
            node_name = GraphRewriterHelper.node_name_from_input(node.name)

            each_node = self.node_details(node=node, outputs=[])

            if node_name not in self.node_name_details:
                self.node_name_details[node_name] = each_node

        for node_name, node_details in self.node_name_details.items():
            # update the upper node's output information.
            for each_input in node_details.node.input:
                self.node_name_details[GraphRewriterHelper.node_name_from_input(each_input)].outputs.append(node_name)

        return self.node_name_details


class GraphRewriterHelper:
    """Encapsulates the graph operation into one class."""

    node_name_cache = {}
    node_name_port_cache = {}

    @staticmethod
    def compare_node_attr(node_a, node_b):
        """Compare two node has identical attributes or not.

        Args:
            node_a (nodedef): Input node.
            node_b (nodedef): Another node to be compared.

        Returns:
            [bool]: True if two node have the identical attributes.
        """
        if len(node_a.input) > 1:
            return False

        if node_a.input != node_b.input:
            return False

        if node_a.op != node_b.op:
            return False

        if len(node_a.attr) != len(node_b.attr):
            return False

        node_a_attr = sorted(list(node_a.attr))
        node_b_attr = sorted(list(node_b.attr))

        if node_a_attr != node_b_attr:
            return False

        for attr_name in node_a_attr:
            if node_a.attr[attr_name] != node_b.attr[attr_name]:
                return False

        return True

    @staticmethod
    def create_node(op, name, inputs):
        """Create a nodedef object.

        Args:
            op (string): op type
            name (string): op name
            inputs (string list): op's inputs name

        Returns:
            nodedef: the created nodedef object
        """
        new_node = node_def_pb2.NodeDef()
        new_node.op = op
        new_node.name = name
        for input_name in inputs:
            new_node.input.extend([input_name])
        return new_node

    @staticmethod
    def create_constant_node(name, value, dtype, shape=None, device="cpu"):
        """Create constant node.

        Args:
            name (string): op name
            value (np.array): input data
            dtype (datatype): data type of the input value
            shape (int list, optional): the value's shape. Defaults to None.
            device (str, optional): the device type, it may be the 'cpu' or 'gpu'.
                                    Defaults to 'cpu'.

        Returns:
            [type]: [description]
        """
        node = GraphRewriterHelper.create_node("Const" if device == "cpu" else "HostConst", name, [])
        GraphRewriterHelper.set_attr_dtype(node, "dtype", dtype)
        GraphRewriterHelper.set_attr_tensor(node, "value", value, dtype, shape)
        return node

    @staticmethod
    def set_attr_dtype(node, key, value):
        """Set the attribute data type."""
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(type=value.as_datatype_enum))

    @staticmethod
    def set_attr_tensor(node, key, value, dtype, shape=None):
        """Set the tensor value to specified attribute field.

        Args:
            node (nodedef): the target nodedef object
            key (string): attribute name
            value (np.array): the content
            dtype (dtypes): data type
            shape (int list, optional): the input tensor's shape. Defaults to None.
        """
        node.attr[key].CopyFrom(
            attr_value_pb2.AttrValue(tensor=tensor_util.make_tensor_proto(value, dtype=dtype, shape=shape))
        )

    @staticmethod
    def set_attr_type_list(node, key, value):
        """Set the node's attr which data type is int list."""
        list_value = attr_value_pb2.AttrValue.ListValue(type=value)
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(list=list_value))

    @staticmethod
    def set_attr_string_list(node, key, value):
        """Set the node's attr which data type is int list."""
        list_value = attr_value_pb2.AttrValue.ListValue(s=value)
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(list=list_value))

    @staticmethod
    def set_attr_string(node, key, value):
        """Set the node's attr which data type is string."""
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(s=value))

    @staticmethod
    def set_attr_int_list(node, key, value):
        """Set the node's attr which data type is int list."""
        list_value = attr_value_pb2.AttrValue.ListValue(i=value)
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(list=list_value))

    @staticmethod
    def set_attr_int(node, key, value):
        """Set the node's attr which data type is int."""
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(i=value))

    @staticmethod
    def set_attr_float(node, key, value):
        """Set the node's attr which data type is float."""
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(f=value))

    @staticmethod
    def set_attr_bool(node, key, value):
        """Set the node's attr which data type is bool."""
        node.attr[key].CopyFrom(attr_value_pb2.AttrValue(b=value))

    @staticmethod
    def node_name_from_input(node_name):
        """Static method that get the valid node name from input name.

        Args:
            node_name (string): node name defined in the input field.

        Returns:
            string: node's name
        """
        if node_name not in GraphRewriterHelper.node_name_cache:
            key = node_name
            if node_name.startswith("^"):
                node_name = node_name[1:]
            m = re.search(r"(.*):\d+$", node_name)
            if m:
                node_name = m.group(1)
            GraphRewriterHelper.node_name_cache[key] = node_name
            return node_name

        return GraphRewriterHelper.node_name_cache[node_name]

    @staticmethod
    def values_from_const(node_def):
        """Extracts the values from a const NodeDef as a numpy ndarray.

        Args:
          node_def: Const NodeDef that has the values we want to access.

        Returns:
          Numpy ndarray containing the values.

        Raises:
          ValueError: If the node isn't a Const.
        """
        assert node_def.op == "Const", "Node named '%s' should be a Const op." % node_def.name

        input_tensor = node_def.attr["value"].tensor
        tensor_value = tensor_util.MakeNdarray(input_tensor)
        return tensor_value

    @staticmethod
    def generate_int32_bias_for_conv(
        bias_tensor,
        channel_size,
        max_input,
        min_input,
        max_filter_tensor,
        min_filter_tensor,
        activation_range,
        weights_range=127.0,
    ):
        """Static method that generate int32 bias for conv op.

        Args:
            bias_tensor: bias node tensor.
            channel_size: channel size.
            max_input: max activation input value.
            min_input: min activation input value.
            max_filter_tensor: max weight input tensor.
            min_filter_tensor: min weight input tensor.
            activation_range: activation range value.
            weights_range: weight range value.

        Returns:
            int32_bias: int32 bias
        """
        bias_length = bias_tensor.shape[0]
        scales = []
        if len(max_filter_tensor) > 1:
            for i in range(channel_size):
                scales.append(
                    activation_range
                    * weights_range
                    / (max(abs(max_input), abs(min_input)) * max(abs(max_filter_tensor[i]), abs(min_filter_tensor[i])))
                )
        else:
            for i in range(channel_size):
                scales.append(
                    activation_range
                    * weights_range
                    / (max(abs(max_input), abs(min_input)) * max(abs(max_filter_tensor[0]), abs(min_filter_tensor[0])))
                )
        int32_bias = []
        if channel_size > 1:
            for i in range(bias_length):
                int32_bias.append((int)(np.around(bias_tensor[i] * scales[i])))
        else:
            for i in range(bias_length):
                int32_bias.append((int)(np.around(bias_tensor[i] * scales[0])))

        return int32_bias

    @staticmethod
    def generate_int32_bias_for_matmul(
        bias_tensor,
        weights_tensor,
        input_range,
        max_input,
        min_input,
        max_filter_value,
        min_filter_value,
    ):
        """Static method that generate int32 bias for matmul op.

        Args:
            bias_tensor: bias node tensor.
            weights_tensor: weights tensor.
            input_range: activation range value.
            max_input: max activation input value.
            min_input: min activation input value.
            max_filter_tensor: max weight input tensor.
            min_filter_tensor: min weight input tensor.

        Returns:
            int32_bias: int32 bias
        """
        bias_scale = 255.0 * 127.0 / (input_range * max(abs(max_filter_value), abs(min_filter_value)))
        relative_scale = 255 * min_input / (max_input - min_input)
        int32_bias = []
        for bias_index, value in enumerate(np.sum(np.array(weights_tensor, dtype=np.int32), axis=0, dtype=np.int32)):
            if bias_index >= bias_tensor.size:
                continue
            int32_bias.append(int(np.around(bias_tensor[bias_index] * bias_scale + value * relative_scale)))

        return int32_bias

    @staticmethod
    def generate_int32_bias_for_matmul_per_channel(
        bias_tensor,
        weights_tensor,
        max_input,
        min_input,
        max_filter_tensor,
        min_filter_tensor,
    ):  # pragma: no cover
        """Static method that generate per-channel int32 bias for matmul op.

        Args:
            bias_tensor: bias node tensor.
            weights_tensor: weights tensor.
            max_input: max activation input value.
            min_input: min activation input value.
            max_filter_tensor: max weight input tensor.
            min_filter_tensor: min weight input tensor.

        Returns:
            int32_bias: int32 bias
        """
        channel_size = bias_tensor.shape[0]
        activation_range = 255.0
        weights_range = 127.0
        scales = []
        relative_scale = 255 * min_input / (max_input - min_input)
        for i in range(channel_size):
            scales.append(
                activation_range
                * weights_range
                / ((max_input - min_input) * max(abs(max_filter_tensor[i]), abs(min_filter_tensor[i])))
            )
        int32_bias = []
        for i in range(channel_size):
            value = np.sum(np.array(weights_tensor), axis=0, dtype=np.int32)[i]
            int32_bias.append((int)(np.around(value * relative_scale + bias_tensor[i] * scales[i])))

        return int32_bias

    @staticmethod
    def gen_valid_sampling_log(log_path):
        """Generate the valid sampling log.

        Args:
          log_path: the valid sampling log file path.

        Returns:
          the sampling min max value.
        """

        def gen_per_iter(data):
            res = []
            requant_tmp = []
            for i in data:
                if i.find("__print__;__requant_") == -1:
                    res.append(i)
                else:
                    requant_tmp.append(i)
            sorted_requant = sorted(requant_tmp)
            odd_list = sorted_requant[::2]
            even_list = sorted_requant[1::2]
            for index, value in enumerate(even_list):
                min_value = min(0, float(value.split(":")[1][1:-1]))
                max_value = float(odd_list[index].split(":")[1][1:-1])
                max_value = max_value if max_value > min_value else min_value + 1e-05
                mixed_str = value.split(":")[0] + "_max:[" + str(min_value) + "][" + str(max_value) + "]"

                res.append(mixed_str)
            return res

        def separate(line):
            """This function is to separate the strings.

            Example:
                ';slice__print__;__max:[1];slice__print__;__min:[-1]' -->
                [';slice__print__;__max:[1]', ';slice__print__;__min:[-1]']
            """
            separated_lines = []
            for subline in line.split("];"):
                if not subline.startswith(";"):
                    subline = ";" + subline
                if not subline.endswith("]"):
                    subline += "]"
                separated_lines.append(subline)
            return separated_lines

        with open(log_path) as f:
            valid_data = []
            for i in f.readlines():
                if not i.startswith(";"):
                    continue
                line = i.strip()
                if line.find("];") != 0:
                    separated_lines = separate(line)
                    valid_data += separated_lines
                else:
                    valid_data.append(line)

        first_line = valid_data[0].rsplit(":")[0]

        iterations = 0
        for i in valid_data:
            if i.startswith(first_line):
                iterations += 1

        step = int(len(valid_data) / iterations)
        if step % 2 == 1:
            step -= 1
            iterations = int(len(valid_data) / step) + int(len(valid_data) % step > 0)

        final_res = []

        for i in range(iterations):
            final_res.extend(gen_per_iter(valid_data[int(i * step) : int(step * (i + 1))]))
            if i + 1 == iterations and int(step * (i + 1)) < len(valid_data):
                final_res.extend(gen_per_iter(valid_data[int(step * (i + 1)) : len(valid_data)]))

        return final_res

    @staticmethod
    def analysis_rnn_model(graph_def, bf16_ops=[], fp32_ops=[]):
        """Match the RNN and dynamic RNN patterns."""
        g = GraphAnalyzer()
        g.graph = graph_def
        graph_info = g.parse_graph()
        rnn_pattern = [["TensorArrayV3"], ["Enter"], ["TensorArrayReadV3"], ["MatMul"], ["BiasAdd"]]
        target_nodes = g.query_fusion_pattern_nodes(rnn_pattern)
        res = {}
        for i in target_nodes:
            if i[-3] not in bf16_ops and i[-3] not in fp32_ops:
                res[(i[-3], i[-2])] = graph_info[i[1]].node.attr["frame_name"].s.decode()

        dynamic_rnn_pattern = [["Enter"], ["MatMul"], ["BiasAdd"]]
        target_nodes = g.query_fusion_pattern_nodes(dynamic_rnn_pattern)
        for i in target_nodes:
            if i[-3] not in bf16_ops and i[-3] not in fp32_ops:
                res[(i[1], i[2])] = graph_info[i[0]].node.attr["frame_name"].s.decode()

        return res



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The configs of algorithms for Keras."""

from __future__ import annotations

from typing import Callable, Dict, List, NamedTuple, Optional, Tuple, Union

import tensorflow as tf

from neural_compressor.common import logger
from neural_compressor.common.base_config import (
    DEFAULT_WHITE_LIST,
    OP_NAME_OR_MODULE_TYPE,
    BaseConfig,
    config_registry,
    register_config,
    register_supported_configs_for_fwk,
)
from neural_compressor.common.utils import STATIC_QUANT

FRAMEWORK_NAME = "keras"


class OperatorConfig(NamedTuple):
    """The config for operator."""

    config: BaseConfig
    operators: List[Union[str, Callable]]
    valid_func_list: List[Callable] = []


@register_config(framework_name=FRAMEWORK_NAME, algo_name=STATIC_QUANT)
class StaticQuantConfig(BaseConfig):
    """Config class for keras static quantization."""

    supported_configs: List[OperatorConfig] = []
    params_list = [
        "weight_dtype",
        "weight_sym",
        "weight_granularity",
        "act_dtype",
        "act_sym",
        "act_granularity",
    ]

    name = STATIC_QUANT

    def __init__(
        self,
        weight_dtype: str = "int8",
        weight_sym: bool = True,
        weight_granularity: str = "per_tensor",
        act_dtype: str = "int8",
        act_sym: bool = True,
        act_granularity: str = "per_tensor",
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
    ):
        """Init static quantization config.

        Args:
            weight_dtype (str): Data type for weights, default is "int".
            weight_sym (bool): Indicates whether weights are symmetric, default is True.
            weight_granularity (str): Calculate tensor-wise scales or channel-wise scales for weights.
            act_dtype (str): Data type for activations, default is "int8".
            act_sym (bool): Indicates whether activations are symmetric, default is True.
            act_granularity (str): Calculate tensor-wise scales or channel-wise scales for activations.
            white_list (list): A list of supported operators of this algorithm.
        """
        super().__init__(white_list=white_list)
        self.weight_dtype = weight_dtype
        self.weight_sym = weight_sym
        self.weight_granularity = weight_granularity
        self.act_dtype = act_dtype
        self.act_sym = act_sym
        self.act_granularity = act_granularity
        self._post_init()

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported configs."""
        supported_configs = []
        static_quant_config = StaticQuantConfig(
            weight_dtype=["int8", "fp32"],
            weight_sym=[True, False],
            weight_granularity=["per_tensor", "per_channel"],
            act_dtype=["int8", "fp32"],
            act_sym=[True, False],
            act_granularity=["per_tensor", "per_channel"],
        )
        operators = [
            tf.keras.layers.Dense,
            tf.keras.layers.Conv2D,
            tf.keras.layers.DepthwiseConv2D,
            tf.keras.layers.SeparableConv2D,
            tf.keras.layers.AvgPool2D,
            tf.keras.layers.MaxPool2D,
            tf.keras.layers.AveragePooling2D,
            tf.keras.layers.MaxPooling2D,
        ]
        supported_configs.append(OperatorConfig(config=static_quant_config, operators=operators))
        cls.supported_configs = supported_configs

    @staticmethod
    def get_model_info(model) -> List[Tuple[str, Callable]]:
        """Get concrete node names for supported operators."""
        white_list = [
            "Dense",
            "Conv2D",
            "DepthwiseConv2D",
            "SeparableConv2D",
            "AvgPool2D",
            "AveragePooling2D",
            "MaxPool2D",
            "MaxPooling2D",
        ]
        filter_result = []

        for layer in model.model.layers:
            if layer.__class__.__name__ in white_list:
                pair = (layer.name, layer.__class__.__name__)
                filter_result.append(pair)
        logger.debug(f"Get model info: {filter_result}")
        return filter_result

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "StaticQuantConfig", List["StaticQuantConfig"]]:
        """Get a default config set for tuning."""
        return StaticQuantConfig(weight_sym=[True, False])


register_supported_configs_for_fwk(fwk_name=FRAMEWORK_NAME)


def get_all_registered_configs() -> Dict[str, BaseConfig]:
    """Get all registered configs for keras framework."""
    registered_configs = config_registry.get_cls_configs()
    return registered_configs.get(FRAMEWORK_NAME, {})


def get_default_static_quant_config() -> StaticQuantConfig:
    """Generate the default static quant config.

    Returns:
        the default keras config.
    """
    return StaticQuantConfig()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""BaseDataloder of all dataloaders."""

import collections
import math
import sys
from abc import abstractmethod

import numpy as np
import tensorflow as tf

from neural_compressor.common import logger


def default_collate(batch):
    """Merge data with outer dimension batch size."""
    elem = batch[0]
    if isinstance(elem, collections.abc.Mapping):
        return {key: default_collate([d[key] for d in batch]) for key in elem}
    elif isinstance(elem, collections.abc.Sequence):
        batch = zip(*batch)
        return [default_collate(samples) for samples in batch]
    elif isinstance(elem, np.ndarray):
        try:
            return np.stack(batch)
        except:
            return batch
    else:
        return batch


class IterableFetcher:
    """Iterate to get next batch-size samples as a batch."""

    def __init__(self, dataset, collate_fn, drop_last, distributed):
        """Initialize IterableFetcher.

        Args:
            dataset (object): dataset object from which to get data
            collate_fn (callable): merge data with outer dimension batch size
            drop_last (bool): whether to drop the last batch if it is incomplete
            distributed (bool): whether the dataloader is distributed
        """
        self.dataset = dataset
        self.collate_fn = collate_fn
        self.drop_last = drop_last
        self.dataset_iter = iter(dataset)
        self.index_whole = 0
        self.process_rank = 0  # The default rank is 0, which represents the main process
        self.process_size = 1  # By default, process_size=1, only the main process is running
        if distributed:
            import horovod.tensorflow as hvd

            hvd.init()
            self.process_rank = hvd.rank()
            self.process_size = hvd.size()
            if self.process_size < 2:
                raise EnvironmentError(
                    "The program is now trying to traverse"
                    " the distributed TensorFlow DefaultDataLoader in only one process."
                    " If you do not want to use distributed DataLoader, please set"
                    " 'distributed: False'. Or If you want to use distributed DataLoader,"
                    " please set 'distributed: True' and launch multiple processes."
                )

    def __call__(self, batched_indices):
        """Fetch data.

        Args:
            batched_indices (list): fetch data according to batched_indices
        """
        batch_data = []
        batch_size = len(batched_indices)
        while True:
            try:
                iter_data = next(self.dataset_iter)
                if (self.index_whole - self.process_rank) % self.process_size == 0:
                    batch_data.append(iter_data)
                self.index_whole += 1
                if len(batch_data) == batch_size:
                    break
            except StopIteration:
                break
        if len(batch_data) == 0 or (self.drop_last and len(batch_data) < len(batched_indices)):
            raise StopIteration
        return self.collate_fn(batch_data)


class IndexFetcher:
    """Take single index or a batch of indices to fetch samples as a batch."""

    def __init__(self, dataset, collate_fn, drop_last, distributed):
        """Initialize IndexFetcher.

        Args:
            dataset (object): dataset object from which to get data
            collate_fn (callable): merge data with outer dimension batch size
            drop_last (bool): whether to drop the last batch if it is incomplete
            distributed (bool): whether the dataloader is distributed
        """
        self.dataset = dataset
        self.collate_fn = collate_fn
        self.drop_last = drop_last

    def __call__(self, batched_indices):
        """Fetch data.

        Args:
            batched_indices (list): fetch data according to batched_indices
        """
        data = [self.dataset[idx] for idx in batched_indices]
        return self.collate_fn(data)


class IterableSampler:
    """Internally samples elements.

    Used for datasets retrieved element by iterator. Yield None to act as a placeholder for each iteration.
    """

    def __init__(self, dataset):
        """Initialize IterableSampler.

        Args:
            dataset (object): dataset object from which to get data
        """
        self.whole_dataset = dataset

    def __iter__(self):
        """Yield data in iterative order."""
        while True:
            yield None

    def __len__(self):
        """Return the length of dataset."""
        return len(self.whole_dataset)


class SequentialSampler:
    """Sequentially samples elements, used for datasets retrieved element by index."""

    def __init__(self, dataset, distributed):
        """Initialize SequentialSampler.

        Args:
            dataset (object): dataset object from which to get data
            distributed (bool): whether the dataloader is distributed
        """
        self.whole_dataset = dataset
        self.distributed = distributed

    def __iter__(self):
        """Yield data in iterative order."""
        self.process_rank = 0  # The default rank is 0, which represents the main process
        self.process_size = 1  # By default, process_size=1, only the main process is running
        if self.distributed:
            import horovod.tensorflow as hvd

            hvd.init()
            self.process_rank = hvd.rank()
            self.process_size = hvd.size()
            if self.process_size < 2:
                raise EnvironmentError(
                    "The program is now trying to traverse"
                    " the distributed TensorFlow DefaultDataLoader in only one process."
                    " If you do not want to use distributed DataLoader, please set"
                    " 'distributed: False'. Or If you want to use distributed DataLoader,"
                    " please set 'distributed: True' and launch multiple processes."
                )
        return iter(range(self.process_rank, len(self.whole_dataset), self.process_size))

    def __len__(self):
        """Return the length of dataset."""
        return len(self.whole_dataset)


class BatchSampler:
    """Yield a batch of indices and number of batches."""

    def __init__(self, sampler, batch_size, drop_last=True):
        """Initialize BatchSampler.

        Args:
            sampler (Sampler): sampler used for generating batches
            batch_size (int): size of batch
            drop_last (bool, optional): whether to drop the last batch if it is incomplete. Defaults to True.
        """
        if isinstance(drop_last, bool):
            self.drop_last = drop_last
        else:
            raise ValueError("last_batch only support bool as input")

        self.sampler = sampler
        self.batch_size = batch_size
        self.drop_last = drop_last

    def __iter__(self):
        """Yield data in iterative order."""
        batch = []
        for idx in self.sampler:
            batch.append(idx)
            if len(batch) == self.batch_size:
                yield batch
                batch = []
        if len(batch) > 0 and not self.drop_last:
            yield batch

    def __len__(self):
        """Return the number of batches."""
        if self.drop_last:
            return len(self.sampler) // self.batch_size
        else:
            return (len(self.sampler) + self.batch_size - 1) // self.batch_size


class BaseDataLoader:
    """Base class for TF DataLoaders.

    _generate_dataloader is needed to create a dataloader object
    from the general params like batch_size and sampler. The dynamic batching is just to
    generate a new dataloader by setting batch_size and last_batch.
    """

    def __init__(
        self,
        dataset,
        batch_size=1,
        last_batch="rollover",
        collate_fn=None,
        sampler=None,
        batch_sampler=None,
        num_workers=0,
        pin_memory=False,
        shuffle=False,
        distributed=False,
    ):
        """Initialize DefaultDataLoader.

        Args:
            dataset (object): dataset from which to load the data
            batch_size (int, optional): number of samples per batch. Defaults to 1.
            last_batch (str, optional): whether to drop the last batch if it is incomplete.
                                        Support ['rollover', 'discard'], rollover means False, discard means True.
                                        Defaults to 'rollover'.
            collate_fn (callable, optional): merge data with outer dimension batch size. Defaults to None.
            sampler (Sampler, optional): Sampler object to sample data. Defaults to None.
            batch_sampler (BatchSampler, optional): BatchSampler object to generate batch of indices. Defaults to None.
            num_workers (int, optional): number of subprocesses to use for data loading. Defaults to 0.
            pin_memory (bool, optional): whether to copy data into pinned memory before returning. Defaults to False.
            shuffle (bool, optional): whether to shuffle data. Defaults to False.
            distributed (bool, optional): whether the dataloader is distributed. Defaults to False.
        """
        self.dataset = dataset
        self.last_batch = last_batch
        self.sampler = sampler
        self.batch_sampler = batch_sampler
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.collate_fn = collate_fn
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.distributed = distributed
        self.drop_last = False if last_batch == "rollover" else True
        if self.collate_fn is None:
            self.collate_fn = default_collate

    def batch(self, batch_size, last_batch="rollover"):
        """Set batch_size and last_batch."""
        self.batch_size = batch_size
        self.last_batch = last_batch

    @property
    def dataloader(self):
        """Return dataloader."""
        return self

    def __iter__(self):
        """Yield data in iterative order."""
        return self._generate_dataloader(
            self.dataset,
            batch_size=self.batch_size,
            last_batch=self.last_batch,
            collate_fn=self.collate_fn,
            sampler=self.sampler,
            batch_sampler=self.batch_sampler,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            shuffle=self.shuffle,
            distributed=self.distributed,
        )

    def __len__(self):
        """Get dataset length."""
        try:
            dataset_len = self.dataset.__len__()
        except (AttributeError, TypeError):
            dataset_len = 0
            for _ in self.dataset:
                dataset_len += 1
        except Exception:
            raise ValueError(
                f"{self.dataset} is invalid, {self.dataset}"
                " does not support calculating the length of its dataloader"
            )
        if self.drop_last is False:
            dataloader_len = math.ceil(dataset_len / self.batch_size)
        else:
            dataloader_len = math.floor(dataset_len / self.batch_size)
        return dataloader_len

    def _generate_dataloader(
        self,
        dataset,
        batch_size,
        last_batch,
        collate_fn,
        sampler,
        batch_sampler,
        num_workers,
        pin_memory,
        shuffle,
        distributed,
    ):
        sampler = self._generate_sampler(dataset, distributed)
        self.batch_sampler = BatchSampler(sampler, batch_size, self.drop_last)

        if self.dataset_type == "index":
            self.fetcher = IndexFetcher(dataset, collate_fn, self.drop_last, distributed)
        elif self.dataset_type == "iter":
            self.fetcher = IterableFetcher(dataset, collate_fn, self.drop_last, distributed)

        for batched_indices in self.batch_sampler:
            try:
                data = self.fetcher(batched_indices)
                yield data
            except StopIteration:
                return

    def _generate_sampler(self, dataset, distributed):
        if hasattr(dataset, "__getitem__"):
            self.dataset_type = "index"
            return SequentialSampler(dataset, distributed)
        elif hasattr(dataset, "__iter__"):
            self.dataset_type = "iter"
            return IterableSampler(dataset)
        else:
            raise ValueError("dataset type only support (index, iter)")


class DummyDataset:  # pragma: no cover
    """Dataset used for dummy data generation.

    This Dataset is to construct a dataset from a specific shape.
    The value range is calculated from: low * stand_normal(0, 1) + high.
    (TODO) construct dummy data from real dataset or iteration of data.
    """

    def __init__(self, shape, low=-128.0, high=127.0, dtype="float32", label=True, transform=None, filter=None):
        """Initialize `DummyDataset` class.

        Args:
            shape (list or tuple): Support create multi shape tensors, use list of tuples
                for each tuple in the list, will create a such size tensor.
            low (list or float, default=-128.): Low out the tensor value range from [0, 1]
                to [0, low] or [low, 0] if low < 0, if float, will implement all tensors with same low value.
            high (list or float, default=127.): High the tensor value by add all tensor element
                value high. If list, length of list should be same with shape list.
            dtype (list or str, default='float32'): Support multi tensor dtype setting.
                If list, length of list should be same with shape list. If str, all tensors will
                use same dtype. dtype supports 'float32', 'float16', 'uint8', 'int8', 'int32', 'int64', 'bool'.
            label (bool, default=True): Whether to return 0 as label.
            transform (transform object, default=None): Dummy dataset does not need transform.
                If transform is not None, it will ignore it.
            filter (Filter objects, default=None): Filter out examples according to specific conditions.
        """
        dtype_map = {
            "float32": np.float32,
            "float16": np.float16,
            "uint8": np.uint8,
            "int8": np.int8,
            "int32": np.int32,
            "int64": np.int64,
            "bool": bool,
            "string": str,
        }

        np.random.seed(9527)
        self.transform = transform
        self.label = label
        if len(shape) == 0:
            logger.info("No data in the dummy dataset.")
        elif isinstance(shape, list):
            # list tensor should same first dimension n
            n = shape[0][0]
            assert all(
                isinstance(elem, tuple) and elem[0] == n for elem in shape
            ), "each tensor shape should be tuple and same first dimension"

            if isinstance(low, list):
                assert len(low) == len(shape) and all(
                    isinstance(elem, float) for elem in low
                ), "low list should have same length with shape with element data type float"
            else:
                low = (low * np.ones(len(shape))).astype(float)

            if isinstance(high, list):
                assert len(high) == len(shape) and all(
                    isinstance(elem, float) for elem in high
                ), "high list should have same length with shape with element data type float"
            else:
                high = (high * np.ones(len(shape))).astype(float)

            if isinstance(dtype, list):
                assert len(dtype) == len(shape) and all(
                    elem in dtype_map.keys() for elem in dtype
                ), "high list should have same length with shape with element data type float"
            else:
                dtype = [dtype for i in range(0, len(shape))]

        elif isinstance(shape, tuple):
            shape = [shape]
            if isinstance(low, float):
                low = [low]
            else:
                assert (
                    isinstance(low, list) and len(low) == 1 and isinstance(low[0], float)
                ), "low should be float or list of float with length 1"

            if isinstance(high, float):
                high = [high]
            else:
                assert (
                    isinstance(high, list) and len(high) == 1 and isinstance(high[0], float)
                ), "high should be float or list of float with length 1"

            if isinstance(dtype, str):
                assert dtype in dtype_map.keys(), "dtype only support {}".format(dtype_map.keys())
                dtype = [dtype]
            else:
                assert (
                    isinstance(dtype, list) and len(dtype) == 1 and dtype[0] in dtype_map.keys()
                ), "dtype should be str or list of str in supported dtypes"

        self.dataset = []
        for idx in range(0, len(shape)):
            tensor = np.random.uniform(low=low[idx], high=high[idx], size=shape[idx])
            tensor = tensor.astype(dtype_map[dtype[idx]])
            self.dataset.append(tensor)

        if len(self.dataset) == 1:
            self.dataset = self.dataset[0]
        else:
            self.dataset = [elem for elem in zip(*self.dataset)]

    def __len__(self):
        """Return the length of dataset."""
        return len(self.dataset)

    def __getitem__(self, index):
        """Return the item of dataset according to the given index."""
        sample = self.dataset[index]
        if self.transform is not None:
            logger.warning("Dummy dataset does not need transform.")

        if self.label:
            return sample, 0
        else:
            return sample


class DummyDatasetV2:  # pragma: no cover
    """Dataset used for dummy_v2 data generation.

    This Dataset is to construct a dataset from a input shape and label shape.
    The value range is calculated from: low * stand_normal(0, 1) + high.
    """

    def __init__(
        self, input_shape, label_shape=None, low=-128.0, high=127.0, dtype="float32", transform=None, filter=None
    ):
        """Initialize `DummyDataset` class.

        Args:
            sample_size (int): Total size of the dummy samples.
            input_shape (list or tuple): Create single or multi input tensors,
                tuple represent the sample shape of the dataset, e.g. an image size should be
                represented as (224, 224, 3), list contains multiple tuple and represent multi input tensors.
            label_shape (list or tuple): Create single or multi label tensors,
                tuple represent the label shape of the dataset, e.g. an label size should be
                represented as (1, ), list contains multiple tuple and represent multi label tensors.
            low (list or float, default=-128.): Low out the tensor value range from [0, 1]
                to [0, low] or [low, 0] if low < 0. If float, will implement all tensors with same low value.
            high (list or float, default=127.): High the tensor value by add all tensor element value high.
                If list, length of list should be same with shape list.
            dtype (list or str, default='float32'): Support multi tensor dtype setting.
                If list, length of list should be same with shape list.
                If str, all tensors will use same dtype.
                dtype supports 'float32', 'float16', 'uint8', 'int8','int32', 'int64', 'bool'.
            transform (transform object, default=None): dummy_v2 dataset does not need transform.
                If transform is not None, it will ignore it.
            filter (Filter objects, default=None): Filter out examples according to specific conditions.
        """
        self.dtype_map = {
            "float32": np.float32,
            "float16": np.float16,
            "uint8": np.uint8,
            "int8": np.int8,
            "int32": np.int32,
            "int64": np.int64,
            "bool": bool,
        }

        np.random.seed(9527)
        self.transform = transform
        self.input_shape = input_shape
        self.label_shape = label_shape
        self.low = low
        self.high = high
        self.dtype = dtype

        if label_shape is None:
            self.label_dim = 0
        elif isinstance(label_shape, tuple):
            self.label_dim = 1
        else:
            self.label_dim = len(label_shape)

        self.input_dim = 1 if isinstance(input_shape, tuple) else len(input_shape)
        self.total_dim = self.input_dim + self.label_dim

        if isinstance(high, list):
            assert len(high) == self.total_dim and all(
                isinstance(elem, float) for elem in high
            ), "high value list length should same with label dim + input_dim"
        else:
            self.high = (high * np.ones(self.total_dim)).astype(np.float32)

        if isinstance(low, list):
            assert len(low) == self.total_dim and all(
                isinstance(elem, float) for elem in low
            ), "low value list length should same with label dim + input_dim"
        else:
            self.low = (low * np.ones(self.total_dim)).astype(np.float32)

        if isinstance(dtype, list):
            assert len(dtype) == self.total_dim and all(
                elem in self.dtype_map.keys() for elem in dtype
            ), "dtype list length should same with label dim + input_dim"
        else:
            self.dtype = [self.dtype for i in range(0, self.total_dim)]

        if isinstance(input_shape, tuple):
            self.input_shape = [input_shape]

        if isinstance(label_shape, tuple):
            self.label_shape = [label_shape]

    def __iter__(self):
        """Yield data in iterative order."""
        while True:
            input_data = []
            for idx in range(0, self.input_dim):
                tensor = np.random.uniform(low=self.low[idx], high=self.high[idx], size=self.input_shape[idx])
                tensor = tensor.astype(self.dtype_map[self.dtype[idx]])
                input_data.append(tensor)

            label = []
            for idx in range(0, self.label_dim):
                shift_idx = self.input_dim + idx
                tensor = np.random.uniform(
                    low=self.low[shift_idx], high=self.high[shift_idx], size=self.label_shape[idx]
                )
                tensor = tensor.astype(self.dtype_map[self.dtype[shift_idx]])
                label.append(tensor)

            if len(input_data) == 1:
                input_data = input_data[0]

            if len(label) == 1:
                label = label[0]

            if len(label) > 0:
                yield input_data, label
            else:
                yield input_data

    def __len__(self):
        """Return the length of dataset."""
        return sys.maxsize



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor TF quantization config API."""


from __future__ import annotations

from enum import Enum
from typing import Callable, Dict, List, NamedTuple, Optional, Tuple, Union

import tensorflow as tf

from neural_compressor.common import logger
from neural_compressor.common.base_config import (
    DEFAULT_WHITE_LIST,
    OP_NAME_OR_MODULE_TYPE,
    BaseConfig,
    config_registry,
    register_config,
    register_supported_configs_for_fwk,
)
from neural_compressor.common.utils import SMOOTH_QUANT, STATIC_QUANT
from neural_compressor.tensorflow.utils import DEFAULT_SQ_ALPHA_ARGS

FRAMEWORK_NAME = "tensorflow"

__all__ = [
    "StaticQuantConfig",
    "get_default_static_quant_config",
    "SmoothQuantConfig",
    "get_default_sq_config",
]


class OperatorConfig(NamedTuple):
    config: BaseConfig
    operators: List[Union[str, Callable]]
    valid_func_list: List[Callable] = []


@register_config(framework_name=FRAMEWORK_NAME, algo_name=STATIC_QUANT)
class StaticQuantConfig(BaseConfig):
    """Config class for tf static quantization."""

    supported_configs: List[OperatorConfig] = []
    params_list = [
        "weight_dtype",
        "weight_sym",
        "weight_granularity",
        "weight_algorithm",
        "act_dtype",
        "act_sym",
        "act_granularity",
        "act_algorithm",
    ]

    name = STATIC_QUANT

    def __init__(
        self,
        weight_dtype: str = "int8",
        weight_sym: bool = True,
        weight_granularity: str = "per_tensor",
        weight_algorithm: str = "minmax",
        act_dtype: str = "int8",
        act_sym: bool = True,
        act_granularity: str = "per_tensor",
        act_algorithm: str = "minmax",
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
    ):
        """Init static quantization config.

        Args:
            weight_dtype (str): Data type for weights, default is "int".
            weight_sym (bool): Indicates whether weights are symmetric, default is True.
            weight_granularity (str): Calculate tensor-wise scales or channel-wise scales for weights.
            weight_algorithm (str): Choose quantization algorithms for weights.
            act_dtype (str): Data type for activations, default is "int8".
            act_sym (bool): Indicates whether activations are symmetric, default is True.
            act_granularity (str): Calculate tensor-wise scales or channel-wise scales for activations.
            act_algorithm (str): Choose quantization algorithms for activations.
            white_list (list): A list of supported operators of this algorithm.
        """
        super().__init__(white_list=white_list)
        self.weight_dtype = weight_dtype
        self.weight_sym = weight_sym
        self.weight_granularity = weight_granularity
        self.weight_algorithm = weight_algorithm
        self.act_dtype = act_dtype
        self.act_sym = act_sym
        self.act_granularity = act_granularity
        self.act_algorithm = act_algorithm
        self._post_init()

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported config."""
        supported_configs = []
        static_quant_config = StaticQuantConfig(
            weight_dtype=["int8", "bf16", "fp32"],
            weight_sym=[True, False],
            weight_granularity=["per_tensor", "per_channel"],
            weight_algorithm=["minmax", "kl"],
            act_dtype=["int8", "bf16", "fp32"],
            act_sym=[True, False],
            act_granularity=["per_tensor"],
            act_algorithm=["minmax", "kl"],
        )
        operators = [
            tf.nn.conv2d,
            tf.raw_ops.FusedBatchNormV3,
            tf.nn.conv3d,
            tf.raw_ops.MatMul,
            tf.raw_ops.BatchMatMul,
            tf.raw_ops.BatchMatMulV2,
            tf.nn.depthwise_conv2d,
            tf.raw_ops.ConcatV2,
            tf.compat.v1.nn.fused_batch_norm,
            tf.nn.max_pool,
            tf.nn.avg_pool,
            tf.compat.v1.nn.conv2d_backprop_input,
            tf.raw_ops.Conv3DBackpropInputV2,
        ]
        supported_configs.append(OperatorConfig(config=static_quant_config, operators=operators))
        cls.supported_configs = supported_configs

    def get_model_info(self, model) -> List[Tuple[str, Callable]]:
        """Get concrete node names for supported operators."""
        white_list = [
            "Conv2D",
            "FusedBatchNormV3",
            "Conv3D",
            "_MklFusedInstanceNorm",
            "MatMul",
            "BatchMatMul",
            "BatchMatMulV2",
            "DepthwiseConv2dNative",
            "ConcatV2",
            "FusedBatchNorm",
            "FusedBatchNormV2",
            "MaxPool",
            "MaxPool3D",
            "AvgPool",
            "Conv2DBackpropInput",
            "Conv3DBackpropInputV2",
        ]
        for key in self._local_config.keys():
            if key in white_list:
                white_list.remove(key)
        filter_result = []
        for node in model.graph_def.node:
            if node.op in white_list:
                pair = (node.name, node.op)
                filter_result.append(pair)
        logger.debug(f"Get model info: {filter_result}")
        return filter_result

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "StaticQuantConfig", List["StaticQuantConfig"]]:
        """Get a default config set for tuning."""
        return StaticQuantConfig(
            weight_dtype=["int8", "fp32"],
            weight_sym=[True, False],
            weight_granularity=["per_tensor", "per_channel"],
            weight_algorithm=["minmax", "kl"],
            act_dtype=["int8", "fp32"],
            act_sym=[True, False],
            act_granularity=["per_tensor"],
            act_algorithm=["minmax", "kl"],
        )


register_supported_configs_for_fwk(fwk_name="keras")


def get_all_registered_configs() -> Dict[str, BaseConfig]:
    """Get all registered configs for keras framework."""
    registered_configs = config_registry.get_cls_configs()
    return registered_configs.get(FRAMEWORK_NAME, {})


def get_default_static_quant_config() -> StaticQuantConfig:
    """Generate the default static quant config.

    Returns:
        the default tf config.
    """
    return StaticQuantConfig()


@register_config(framework_name=FRAMEWORK_NAME, algo_name=SMOOTH_QUANT)
class SmoothQuantConfig(BaseConfig):
    """Config class for tf smooth quantization."""

    supported_configs: List[OperatorConfig] = []
    params_list = [
        "alpha",
        "folding",
        "percentile",
        "op_types",
        "scales_per_op",
        "record_max_info",
        "weight_clip",
        "auto_alpha_args",
    ]
    name = SMOOTH_QUANT

    def __init__(
        self,
        alpha: float = 0.5,
        folding: bool = False,
        percentile: float = 99.999,
        op_types: list = ["MatMul", "Conv2D"],
        scales_per_op: bool = True,
        record_max_info: bool = False,
        weight_clip: bool = True,
        auto_alpha_args: Dict = DEFAULT_SQ_ALPHA_ARGS,
        white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST,
    ):
        """Init RTN weight-only quantization config.

        Args:
            alpha (float): Value to balance input and weight quantization error, between 0 and 1, default is 0.5.
            folding (bool): Whether to fold mul into the previous layer, default is False.
            percentile (float): percentile of calibration to remove outliers, default is 99.99.
            op_types (list): The op types whose input tensor will be dumped, default is ["MatMul", "Conv2D"].
            scales_per_op (bool): Whether to set individual scale for every op, default is True.
            record_max_info (bool): whether record the max info in model for alpha tuning, default is False.
            weight_clip: Whether to clip weight when calculating scales, default is True.
            auto_alpha_args(dict) : Hyperparameters used to set the alpha search space in SQ auto-tuning,
                                    by default the search space is 0.0-1.0 with step_size 0.1.
            white_list (list): A list of supported operators of this algorithm.
        """
        super().__init__()
        self.alpha = alpha
        self.folding = folding
        self.percentile = percentile
        self.op_types = op_types
        self.scales_per_op = scales_per_op
        self.record_max_info = record_max_info
        self.weight_clip = weight_clip
        self.auto_alpha_args = auto_alpha_args
        self.white_list = white_list
        self._post_init()

    @classmethod
    def register_supported_configs(cls) -> List[OperatorConfig]:
        """Register supported configs."""
        supported_configs = []
        smooth_quant_config = SmoothQuantConfig()
        operators = ["MatMul", "Conv2D"]
        supported_configs.append(OperatorConfig(config=smooth_quant_config, operators=operators))
        cls.supported_configs = supported_configs

    @staticmethod
    def get_model_info(model) -> List[Tuple[str, Callable]]:
        """Get concrete node names for supported operators."""
        white_list = ["MatMul", "Conv2D"]
        filter_result = []
        for node in model.graph_def.node:
            if node.op in white_list:
                pair = (node.name, node.op)
                filter_result.append(pair)
        logger.debug(f"Get model info: {filter_result}")
        return filter_result

    @classmethod
    def get_config_set_for_tuning(cls) -> Union[None, "SmoothQuantConfig", List["SmoothQuantConfig"]]:
        """Get a default config set for tuning."""
        return SmoothQuantConfig(alpha=0.5)


SmoothQuantConfig.register_supported_configs()


def get_default_sq_config() -> SmoothQuantConfig:
    """Generate the default rtn config.

    Returns:
        the default smooth quant config.
    """
    return SmoothQuantConfig()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuse Pad into Conv Graph Rewriter."""

import tensorflow as tf
from tensorflow.python.framework import tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_rewriter.graph_base import GraphRewriterBase
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import version1_gt_version2


class FusePadWithConv2DOptimizer(GraphRewriterBase):
    """Fuse Pad op into Conv2D/DepthwiseConv2dNative/Conv3D."""

    def __init__(self, model, excluded_op_names, inputs, cfg, new_api, itex_qdq_mode=False):
        """Initialization."""
        super().__init__(model)
        self.excluded_conv = excluded_op_names
        self.inputs = inputs
        self.cfg = cfg
        self.new_api = new_api
        self.itex_qdq_mode = itex_qdq_mode

    def do_transformation(self):
        """Fuse Pad + Conv2D/DepthwiseConv2dNative/Conv3D --> Conv2D/DepthwiseConv2dNative/Conv3D."""
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model

        graph_info = cur_graph.parse_graph()

        target_nodes = cur_graph.query_fusion_pattern_nodes(
            [["Pad"], ["Conv2D", "Conv3D", "DepthwiseConv2dNative"], ("BiasAdd", "Add", "AddV2")]
        )

        padding_tensor_dict = {}
        for node_combination in target_nodes:
            conv_name = node_combination[1]

            pattern = node_combination[-1]

            if conv_name not in self.cfg:
                continue

            is_perchannel = self.cfg[conv_name][0]

            # Line 55 to line 65 should be removed once the TFDO enabling the single quantized
            # conv2D supporting.
            if len(pattern) == 2:
                # TODO we need to enable single quantizedconv2d with s8 input.
                if not is_perchannel and not cur_graph.has_positive_input(conv_name):
                    continue
                # TFDO has the limitation that the single QuantizedConv2DPerchannel doesn't
                # support padding_list filed.
                if is_perchannel:
                    continue

            if conv_name in self.excluded_conv:
                continue

            padding_tensor = None
            pad_node = None
            if node_combination[0] not in padding_tensor_dict:
                pad_node = graph_info[node_combination[0]].node
                if graph_info[pad_node.input[1]].node.op != "Const":
                    input_node = graph_info[pad_node.input[1]].node
                    if input_node.op == "DataFormatVecPermute":
                        parent_input_node = graph_info[input_node.input[0]].node
                        if parent_input_node.op == "Const":
                            padding_tensor = tensor_util.MakeNdarray(parent_input_node.attr["value"].tensor).flatten()
                        else:
                            continue
                    else:
                        continue
                else:
                    padding_tensor = tensor_util.MakeNdarray(
                        graph_info[pad_node.input[1]].node.attr["value"].tensor
                    ).flatten()
                padding_tensor_dict[node_combination[0]] = padding_tensor
            else:
                padding_tensor = padding_tensor_dict[node_combination[0]]

            if self.itex_qdq_mode:
                enabled_pad_conv2d = bool(
                    tf.version.VERSION == "1.15.0-up3" or version1_gt_version2(tf.version.VERSION, "2.7")
                )
            else:
                enabled_pad_conv2d = bool(tf.version.VERSION == "1.15.0-up3" or self.new_api)

            if any(padding_tensor) and not enabled_pad_conv2d:  # pragma: no cover
                continue

            if pad_node:
                if graph_info[pad_node.input[1]].node.op != "Const":
                    cur_graph.node_name_details[pad_node.name].node.input.remove(pad_node.input[1])
                    cur_graph.remove_node_with_single_input_output(pad_node.name)
                else:
                    cur_graph.remove_node_with_single_input_output(pad_node.name)
                    cur_graph.remove_node(pad_node.input[1])
            conv_node = graph_info[node_combination[1]].node
            if self.itex_qdq_mode:
                if any(padding_tensor) and enabled_pad_conv2d:  # pragma: no cover
                    Helper.set_attr_string(conv_node, "padding", b"EXPLICIT")
                    Helper.set_attr_int_list(conv_node, "explicit_paddings", padding_tensor)
            else:
                Helper.set_attr_int_list(conv_node, "padding_list", padding_tensor)
                if any(padding_tensor) and enabled_pad_conv2d:  # pragma: no cover
                    Helper.set_attr_string(conv_node, "padding", b"EXPLICIT")

        return cur_graph.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuse Conv with Math Graph Rewriter."""

import numpy as np
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class FuseConvWithMathOptimizer(GraphRewriterBase):
    """Convert below subgraph to Conv2D + BiasAdd by eliminating math ops.

    Conv2D            Conv2D
      |                  |
      Sub                 |
      |       ---->      |
    RealDiv              |
      |                  |
      Mul                 |
      |                  |
    BiasAdd           BiasAdd
    """

    @dump_elapsed_time("Pass FuseConvWithMathOptimizer")
    def do_transformation(self):
        """Fuse Conv + Sub + RealDiv + Mul + BiasAdd to Conv + BiasAdd."""
        g = GraphAnalyzer()
        g.graph = self.model
        graph_info = g.parse_graph()
        pattern_definition = [["Conv2D"], ["Sub"], ["RealDiv"], ["Mul"], ["BiasAdd"]]
        target_nodes = g.query_fusion_pattern_nodes(pattern_definition)
        for i in target_nodes:
            weights_node_name = graph_info[i[0]].node.input[1]
            weights_node = graph_info[weights_node_name].node

            sub_input_names = list(graph_info[i[1]].node.input)
            sub_content_node_name = list(set(sub_input_names).difference([i[0]]))[0]
            sub_content_node = graph_info[sub_content_node_name].node
            sub_tensor = tensor_util.MakeNdarray(sub_content_node.attr["value"].tensor)

            real_div_input_names = list(graph_info[i[2]].node.input)
            real_div_content_node_name = list(set(real_div_input_names).difference([i[1]]))[0]
            real_div_node = graph_info[real_div_content_node_name].node
            real_div_tensor = tensor_util.MakeNdarray(real_div_node.attr["value"].tensor)

            mul_input_names = list(graph_info[i[3]].node.input)
            mul_content_node_name = list(set(mul_input_names).difference([i[2]]))[0]
            mul_content_node = graph_info[mul_content_node_name].node
            mul_tensor = tensor_util.MakeNdarray(mul_content_node.attr["value"].tensor)

            bias_input_names = list(graph_info[i[4]].node.input)
            bias_content_node_name = list(set(bias_input_names).difference([i[3]]))[0]
            bias_content_node = graph_info[bias_content_node_name].node
            bias_tensor = tensor_util.MakeNdarray(bias_content_node.attr["value"].tensor)

            bias_offset_value = bias_tensor - sub_tensor * mul_tensor / real_div_tensor
            weights_offset = mul_tensor / real_div_tensor

            weights = Helper.values_from_const(weights_node)
            original_shape = weights.shape
            tmp_shape = (original_shape[-1], int(weights.size / original_shape[-1]))
            tmp_order = [weights.ndim - 1] + [i for i in range(weights.ndim - 1)]

            scaled_weights = np.copy(weights).transpose(tmp_order).ravel().reshape(tmp_shape)
            reshape_scale = np.array(weights_offset).reshape(len(weights_offset), 1)
            scaled_weights = np.multiply(scaled_weights, reshape_scale).transpose().reshape(original_shape)
            scaled_weight_name = weights_node_name + "_conv_math_offset"
            scaled_weights_node = Helper.create_constant_node(
                scaled_weight_name, scaled_weights, dtypes.float32, shape=weights.shape
            )

            g.add_node(scaled_weights_node, None, [i[0]])
            g.replace_const_node(scaled_weights_node, [i[0]], weights_node_name)

            offset_node = Helper.create_constant_node(i[0] + "_biasadd_math_offset", bias_offset_value, dtypes.float32)
            g.add_node(offset_node, None, [i[4]])
            graph_info[i[4]].node.input[0] = i[0]

            graph_info[i[4]].node.input[1] = offset_node.name

            g.remove_node(i[1])
            g.remove_node(sub_content_node_name)

            g.remove_node(i[2])
            g.remove_node(real_div_content_node_name)

            g.remove_node(i[3])
            g.remove_node(mul_content_node_name)

            g.remove_node(bias_content_node_name)

        return g.dump_graph()



# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The auto-tune module."""

import copy
import uuid
from typing import Any, Callable, Dict, Generator, Iterator, List, Optional, Sized, Tuple, Union

from neural_compressor.common.base_config import BaseConfig
from neural_compressor.common.utils import TuningLogger, logger

__all__ = [
    "Evaluator",
    "EvaluationFuncWrapper",
    "TuningConfig",
    "Sampler",
    "ConfigLoader",
    "TuningMonitor",
    "init_tuning",
    "Sampler",
    "SequentialSampler",
    "default_sampler",
    "ConfigSet",
]


class EvaluationFuncWrapper:
    """Evaluation function wrapper."""

    def __init__(self, eval_fn: Callable, eval_args=None):
        """Evaluation function wrapper.

        Args:
            eval_fn: a function for evaluated the float or quantized model
            eval_args: positional arguments for `eval_fn`
        """
        self.eval_fn = eval_fn
        self.eval_args = eval_args

    def evaluate(self, model) -> Union[float, int]:
        """Evaluates the given model using the evaluation function and arguments provided.

        Args:
            model: The model to be evaluated.

        Returns:
            The evaluation result, which can be a float or an integer.
        """
        result = self.eval_fn(model, *self.eval_args) if self.eval_args else self.eval_fn(model)
        return result


class Evaluator:
    """Evaluator is a collection of evaluation functions.

    Note: will deprecate this class in the future.

    Examples:
        def eval_acc(model):
            ...

        def eval_perf(molde):
            ...

        # Usage
        user_eval_fns1 = eval_acc
        user_eval_fns2 = {"eval_fn": eval_acc}
        user_eval_fns3 = {"eval_fn": eval_acc, "weight": 1.0, "name": "accuracy"}
        user_eval_fns4 = [
            {"eval_fn": eval_acc, "weight": 0.5},
            {"eval_fn": eval_perf, "weight": 0.5, "name": "accuracy"},
            ]
    """

    EVAL_FN = "eval_fn"
    WEIGHT = "weight"
    FN_NAME = "name"
    EVAL_FN_TEMPLATE: Dict[str, Any] = {EVAL_FN: None, WEIGHT: 1.0, FN_NAME: None}

    def __init__(self) -> None:
        """Initializes the BaseTuning class."""
        self.eval_fn_registry: List[Dict[str, Any]] = []

    def evaluate(self, model) -> float:
        """Evaluate the model using registered evaluation functions.

        Args:
            model: The fp32 model or quantized model.

        Returns:
            The overall result of all registered evaluation functions.
        """
        result = 0
        for eval_pair in self.eval_fn_registry:
            eval_fn = eval_pair[self.EVAL_FN]
            eval_result = eval_fn(model)
            result = self._update_the_objective_score(eval_pair, eval_result, result)
        return result

    def _update_the_objective_score(self, eval_pair, eval_result, overall_result) -> float:
        return overall_result + eval_result * eval_pair[self.WEIGHT]

    def get_number_of_eval_functions(self) -> int:
        """Returns the number of evaluation functions in the eval_fn_registry.

        Returns:
            int: The number of evaluation functions.
        """
        return len(self.eval_fn_registry)

    def _set_eval_fn_registry(self, user_eval_fns: List[Dict]) -> None:
        self.eval_fn_registry = [
            {
                self.EVAL_FN: user_eval_fn_pair[self.EVAL_FN],
                self.WEIGHT: user_eval_fn_pair.get(self.WEIGHT, 1.0),
                self.FN_NAME: user_eval_fn_pair.get(self.FN_NAME, user_eval_fn_pair[self.EVAL_FN].__name__),
            }
            for user_eval_fn_pair in user_eval_fns
        ]

    def set_eval_fn_registry(self, eval_fns: Optional[Union[Callable, Dict, List[Dict]]] = None) -> None:
        """Set the evaluation function registry.

        Args:
            eval_fns (Optional[Union[Callable, Dict, List[Dict]]]): The evaluation function(s) to be registered.
                It can be a single function, a dictionary, or a list of dictionaries.
                If `eval_fns` is None, the method returns without making any changes.
                If `eval_fns` is a single function, it will be converted into a dictionary and added to the registry.
                If `eval_fns` is a dictionary, it will be added to the registry as is.
                If `eval_fns` is a list of dictionaries, each dictionary will be added to the registry.

        Raises:
            NotImplementedError: If `eval_fns` is not a dict or a list of dicts.

        Note:
            The format of the evaluation function(s) should follow the class docstring for details.
        """
        if eval_fns is None:
            return
        elif callable(eval_fns):
            # single eval_fn
            eval_fn_pair = copy.deepcopy(self.EVAL_FN_TEMPLATE)
            eval_fn_pair[self.EVAL_FN] = eval_fns
            eval_fn_pair[self.FN_NAME] = eval_fns.__name__
            eval_fns = [eval_fn_pair]
        elif isinstance(eval_fns, Dict):
            eval_fns = [eval_fns]
        elif isinstance(eval_fns, List):
            assert all([isinstance(eval_fn_pair, Dict) for eval_fn_pair in eval_fns])
        else:
            raise NotImplementedError(f"The eval_fns should be a dict or a list of dict, but got {type(eval_fns)}.")
        self._set_eval_fn_registry(eval_fns)

    def self_check(self) -> None:
        """Perform a self-check to ensure that there is at least one evaluation metric registered for auto-tune.

        Raises:
            AssertionError: If no evaluation metric is registered for auto-tune.
        """
        # check the number of evaluation functions
        num_eval_fns = self.get_number_of_eval_functions()
        assert num_eval_fns > 0, "Please ensure that you register at least one evaluation metric for auto-tune."
        logger.info("There are %d evaluations functions.", num_eval_fns)


evaluator = Evaluator()


class ConfigSet:
    """A class representing a set of configurations.

    Args:
        config_list (List[BaseConfig]): A list of BaseConfig objects.

    Attributes:
        config_list (List[BaseConfig]): The list of BaseConfig objects.
    """

    def __init__(self, config_list: List[BaseConfig]) -> None:
        """Initializes a ConfigSet object."""
        self.config_list = config_list

    def __getitem__(self, index) -> BaseConfig:
        """Get the config object by index."""
        assert 0 <= index < len(self.config_list), f"Index {index} out of range."
        return self.config_list[index]

    def __len__(self) -> int:
        """Get the number of configs in the config_list."""
        return len(self.config_list)

    @classmethod
    def _from_single_config(cls, config: BaseConfig) -> List[BaseConfig]:
        config_list = []
        config_list = config.expand()
        return config_list

    @classmethod
    def _from_list_of_configs(cls, fwk_configs: List[BaseConfig]) -> List[BaseConfig]:
        config_list = []
        for config in fwk_configs:
            config_list += cls._from_single_config(config)
        return config_list

    @classmethod
    def generate_config_list(cls, fwk_configs: Union[BaseConfig, List[BaseConfig]]):
        """Generate the config_list based on the input fwk_configs.

        There are several cases for the input `fwk_configs`:
            1. fwk_configs is a single config
            2. fwk_configs is a list of configs
            For a single config, we need to check if it can be expanded or not.

        Args:
            fwk_configs (Union[BaseConfig, List[BaseConfig]]): A single config or a list of configs.

        Returns:
            List[BaseConfig]: The generated config_list.
        """
        config_list = []
        if isinstance(fwk_configs, BaseConfig):
            config_list = cls._from_single_config(fwk_configs)
        elif isinstance(fwk_configs, List):
            config_list = cls._from_list_of_configs(fwk_configs)
        else:
            raise NotImplementedError(f"Unsupported type {type(fwk_configs)} for fwk_configs.")
        return config_list

    @classmethod
    def from_fwk_configs(cls, fwk_configs: Union[BaseConfig, List[BaseConfig]]) -> "ConfigSet":
        """Create a ConfigSet object from a single config or a list of configs.

        Args:
            fwk_configs: A single config or a list of configs.

        Examples of `fwk_configs`:
            1) single config: RTNConfig(weight_group_size=32)
            2) single expandable config: RTNConfig(weight_group_size=[32, 64])
            3) mixed 1) and 2): [RTNConfig(weight_group_size=32), RTNConfig(weight_group_size=[32, 64])]

        Returns:
            ConfigSet: A ConfigSet object.
        """
        config_list = cls.generate_config_list(fwk_configs)
        return cls(config_list)


class Sampler:
    """Base class for samplers."""

    def __init__(self, config_source: Optional[ConfigSet]) -> None:
        """Initializes a Sampler object."""
        pass

    def __iter__(self) -> Iterator[BaseConfig]:
        """Iterate over indices of config set elements."""
        raise NotImplementedError


class SequentialSampler(Sampler):
    """Samples elements sequentially, always in the same order.

    Args:
        config_source (_ConfigSet): config set to sample from
    """

    config_source: Sized

    def __init__(self, config_source: Sized) -> None:
        """Initializes a SequentialSampler object."""
        self.config_source = config_source

    def __iter__(self) -> Iterator[int]:
        """Iterate over indices of config set elements."""
        return iter(range(len(self.config_source)))

    def __len__(self) -> int:
        """Get the number of configs in the config_source."""
        return len(self.config_source)


default_sampler = SequentialSampler


class ConfigLoader:
    """ConfigLoader is a generator that yields configs from a config set."""

    def __init__(
        self, config_set: ConfigSet, sampler: Sampler = default_sampler, skip_verified_config: bool = True
    ) -> None:
        """Initializes the ConfigLoader class.

        Args:
            config_set (ConfigSet): The configuration set.
            sampler (Sampler, optional): The sampler to use for sampling configurations. Defaults to default_sampler.
            skip_verified_config (bool, optional): Whether to skip verified configurations. Defaults to True.
        """
        self.config_set = ConfigSet.from_fwk_configs(config_set)
        self._sampler = sampler(self.config_set)
        self.skip_verified_config = skip_verified_config
        self.verify_config_list = list()

    def is_verified_config(self, config):
        """Check if the config is verified."""
        for verified_config in self.verify_config_list:
            if config == verified_config:
                return True
        return False

    def __iter__(self) -> Generator[BaseConfig, Any, None]:
        """Iterate over the config set and yield configs."""
        for index in self._sampler:
            new_config = self.config_set[index]
            if self.skip_verified_config and self.is_verified_config(new_config):
                logger.warning("Skip the verified config:")
                logger.warning(new_config.to_dict())
                continue
            self.verify_config_list.append(new_config)
            yield new_config


class TuningConfig:
    """Config for auto tuning pipeline.

    Examples:
        from neural_compressor.torch.quantization import TuningConfig
        tune_config = TuningConfig(
            config_set=[config1, config2, ...],
            max_trials=3,
            tolerable_loss=0.01)

        The tuning process stops when either of the following conditions is met:
            1) The number of trials reaches the maximum trials.
            2) The metric loss is within the tolerable loss.

        For condition 2), we calculate the metric loss as follows:
            relative_loss = (fp32_baseline - eval_result_of_q_model) / fp32_baseline
            If relative_loss <= tolerable_loss, we stop the tuning process.
            For example:
                tolerable_loss = 0.01
                fp32_baseline = 100
                eval_result_of_q_model = 99
                relative_loss = (100 - 99) / 100 = 0.01
                The metric loss is within the tolerable loss, so the tuning process is stopped.
    """

    def __init__(
        self,
        config_set: Union[BaseConfig, List[BaseConfig]] = None,
        sampler: Sampler = default_sampler,
        tolerable_loss=0.01,
        max_trials=100,
    ):
        """Initial a TuningConfig.

        Args:
            config_set: A single config or a list of configs. Defaults to None.
            sampler: tuning sampler that decide the trials order. Defaults to default_sampler.
            tolerable_loss: This float indicates how much metric loss we can accept.
                The metric loss is relative, it can be both positive and negative. Default is 0.01.
            max_trials: Max tuning times. Combine with `tolerable_loss` field to decide when to stop. Default is 100.
        """
        self.config_set = config_set
        self.sampler = sampler
        self.tolerable_loss = tolerable_loss
        self.max_trials = max_trials


class _TrialRecord:
    @staticmethod
    def _generate_unique_id():
        unique_id = str(uuid.uuid4())
        return unique_id

    def __init__(self, trial_index: int, trial_result: Union[int, float], quant_config: BaseConfig):
        # The unique id to refer to one trial
        self.trial_id = _TrialRecord._generate_unique_id()
        self.trial_index = trial_index
        self.trial_result = trial_result
        self.quant_config = quant_config


class TuningMonitor:
    """The tuning monitor class for auto-tuning."""

    def __init__(self, tuning_config: TuningConfig) -> None:
        """Initialize a TuningMonitor class.

        Args:
            tuning_config (TuningConfig): The configuration object for tuning.

        Attributes:
            tuning_config (TuningConfig): The configuration object for tuning.
            trial_cnt (int): The number of trials performed.
            tuning_history (List[_TrialRecord]): The history of tuning records.
            baseline: The baseline value for comparison.
        """
        self.tuning_config = tuning_config
        self.trial_cnt = 0
        self.tuning_history: List[_TrialRecord] = []
        self.baseline = None

    def add_trial_result(self, trial_index: int, trial_result: Union[int, float], quant_config: BaseConfig) -> None:
        """Adds a trial result to the tuning history.

        Args:
            trial_index (int): The index of the trial.
            trial_result (Union[int, float]): The result of the trial.
            quant_config (BaseConfig): The quantization configuration used for the trial.
        """
        self.trial_cnt += 1
        trial_record = _TrialRecord(trial_index, trial_result, quant_config)
        self.tuning_history.append(trial_record)

    def set_baseline(self, baseline: float):
        """Set the baseline value for auto-tune.

        Args:
            baseline (float): The baseline value to be set.
        """
        self.baseline = baseline
        logger.info(f"Fp32 baseline is {self.baseline}")

    def get_number_of_trials(self):
        """Returns the number of trials in the tuning history."""
        return len(self.tuning_history)

    def get_best_trial_record(self) -> _TrialRecord:
        """Returns the best trial record based on the trial result.

        Raises:
            AssertionError: If there are no trial records in the tuning monitor.

        Returns:
            The best trial record.
        """
        assert self.get_number_of_trials() > 0, "No trial record in tuning monitor."
        # Put the record with a higher score at the beginning
        sorted_trials_records: List[_TrialRecord] = sorted(
            self.tuning_history, key=lambda x: x.trial_result, reverse=True
        )
        return sorted_trials_records[0]

    def get_best_quant_config(self) -> BaseConfig:
        """Get the best quantization configuration based on the best trial record.

        Returns:
            The best quantization configuration (BaseConfig).
        """
        best_trial_record = self.get_best_trial_record()
        return best_trial_record.quant_config

    def need_stop(self) -> bool:
        """Check if need to stop tuning. Either accuracy goal is met, max trials is reached or timeout is reached.

        Returns:
            stop_flag: True if need to stop, otherwise False.
        """
        # reach max trials
        reach_max_trials = self.trial_cnt >= self.tuning_config.max_trials
        # reach accuracy goal
        meet_accuracy_goal = (
            False
            if self.baseline is None
            else self.tuning_history[-1].trial_result >= (self.baseline * (1 - self.tuning_config.tolerable_loss))
        )
        # [-1] is the last element representing the latest trail record.
        return reach_max_trials or meet_accuracy_goal


def init_tuning(tuning_config: TuningConfig) -> Tuple[ConfigLoader, TuningLogger, TuningMonitor]:
    """Initializes the tuning process.

    Args:
        tuning_config (TuningConfig): The configuration for the tuning process.
    """
    config_loader = ConfigLoader(config_set=tuning_config.config_set, sampler=tuning_config.sampler)
    tuning_logger = TuningLogger()
    tuning_monitor = TuningMonitor(tuning_config)
    return config_loader, tuning_logger, tuning_monitor



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuse QuantizedConv Requantize/Dequantize Graph Rewriter."""

import tensorflow as tf
from tensorflow.core.framework import attr_value_pb2, node_def_pb2
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper

from ..graph_base import GraphRewriterBase


class FuseConvRequantizeTransformer(GraphRewriterBase):
    """Fuse Quantized Conv Op with the successor Requantize Op."""

    fuse_patterns = [
        [
            "QuantizedConv2DWithBiasAndRelu",
            "QuantizedDepthwiseConv2DWithBiasAndRelu",
            "QuantizedConv2DWithBias",
            "QuantizedDepthwiseConv2DWithBias",
            "_FusedQuantizedConv2D",
            "_FusedQuantizedDepthwiseConv2D",
            "_FusedQuantizedConv3D",
            "_FusedQuantizedDeconv2D",
            "_FusedQuantizedDeconv3D",
        ],
        ["RequantizePerChannel", "Requantize"],
        ("Dequantize",),
    ]

    fuse_sum_op_types = (
        [b"BiasAdd", b"Sum"],
        [b"BiasAdd", b"Sum", b"Relu"],
        [b"BiasAdd", b"Sum", b"LeakyRelu"],
        [b"BiasAdd", b"Relu", b"Sum"],
        [b"BiasAdd", b"LeakyRelu", b"Sum"],
    )

    sum_pattern = [
        [
            "QuantizedConv2DWithBiasSumAndRelu",
            "QuantizedConv2DWithBiasReluAndSum",
            "_FusedQuantizedDepthwiseConv2D",
            "_FusedQuantizedConv2D",
            "_FusedQuantizedConv3D",
        ],
        ["RequantizePerChannel", "Requantize"],
    ]

    def __init__(self, model, device="cpu", new_api=False):
        """Initialization."""
        super().__init__(model)
        self.device = device
        self.graph_analyzer = GraphAnalyzer()
        self.graph_analyzer.graph = self.model
        self.graph_info = self.graph_analyzer.parse_graph()
        self.fused_ops = []
        self.output_types = []
        self.new_api = new_api

    def do_transformation(self):
        """Fuse the quantized op with the following requantize op.

        The transformation has two stages, the first step is to fuse the patterns
        defined in self.fuse_patterns and the last step is to fuse the self.sum_patterns.

        Returns:
            [graphdef]: the optimized graphdef object
        """
        int8_type = dtypes.qint8.as_datatype_enum
        uint8_type = dtypes.quint8.as_datatype_enum
        float32_type = dtypes.float32.as_datatype_enum
        qint32_type = dtypes.qint32.as_datatype_enum
        dtype_map_dict = {
            dtypes.qint8.as_datatype_enum: dtypes.qint8,
            dtypes.quint8.as_datatype_enum: dtypes.quint8,
            dtypes.float32.as_datatype_enum: dtypes.float32,
            dtypes.qint32.as_datatype_enum: dtypes.qint32,
        }
        target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(self.fuse_patterns)

        for i in target_nodes:
            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            if not self.new_api and quantized_node.op == "QuantizedDepthwiseConv2DWithBias":
                continue
            if i[-1][0] in (
                "_FusedQuantizedDepthwiseConv2D",
                "_FusedQuantizedConv2D",
                "_FusedQuantizedConv3D",
                "_FusedQuantizedDeconv2D",
                "_FusedQuantizedDeconv3D",
            ):
                if str(quantized_node.attr["fused_ops"].list.s).find("Sum") != -1:
                    continue
                # else:
                #   print(quantized_node.attr['fused_ops'].list.s)
            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            if (
                i[-1][-1] == "Dequantize"
                and self.new_api
                and i[0]
                in (
                    "_FusedQuantizedDepthwiseConv2D",
                    "_FusedQuantizedConv2D",
                    "_FusedQuantizedConv3D",
                    "_FusedQuantizedDeconv2D",
                    "_FusedQuantizedDeconv3D",
                )
            ):
                dequantize_node_name = i[2]
            else:
                dequantize_node_name = None
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]

            quantized_node_op = i[-1][0]

            new_node = node_def_pb2.NodeDef()
            if self.new_api:
                if i[-1][0] == "QuantizedConv2DWithBiasAndRelu":
                    new_node.op = "_FusedQuantizedConv2D"
                    self.fused_ops = [b"BiasAdd", b"Relu", b"Requantize"]
                    self.output_types = [
                        requantize_node.attr["out_type"].type,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ]
                elif i[-1][0] == "QuantizedConv2DWithBias":
                    new_node.op = "_FusedQuantizedConv2D"
                    self.fused_ops = [b"BiasAdd", b"Requantize"]
                    self.output_types = [
                        requantize_node.attr["out_type"].type,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ]
                elif i[-1][0] == "QuantizedDepthwiseConv2DWithBias":
                    new_node.op = "_FusedQuantizedDepthwiseConv2D"
                    self.fused_ops = [b"BiasAdd", b"Requantize"]
                    self.output_types = [
                        requantize_node.attr["out_type"].type,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ]
                elif i[-1][0] == "QuantizedDepthwiseConv2DWithBiasAndRelu":
                    new_node.op = "_FusedQuantizedDepthwiseConv2D"
                    self.fused_ops = [b"BiasAdd", b"Relu", b"Requantize"]
                    self.output_types = [
                        requantize_node.attr["out_type"].type,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ]
                elif quantized_node_op == "_FusedQuantizedConv2D":
                    new_node.op = "_FusedQuantizedConv2D"
                elif quantized_node_op == "_FusedQuantizedDepthwiseConv2D":
                    new_node.op = "_FusedQuantizedDepthwiseConv2D"
                elif quantized_node_op == "_FusedQuantizedConv3D":
                    new_node.op = "_FusedQuantizedConv3D"
                elif quantized_node_op == "_FusedQuantizedDeconv2D":
                    new_node.op = "_FusedQuantizedDeconv2D"
                elif quantized_node_op == "_FusedQuantizedDeconv3D":
                    new_node.op = "_FusedQuantizedDeconv3D"
            else:
                new_node.op = quantized_node_op + "AndRequantize"
            new_node.name = requantize_node_name
            for _, value in enumerate(quantized_node.input):
                new_node.input.append(value)

            if "Tinput" in quantized_node.attr:
                new_node.attr["Tinput"].CopyFrom(quantized_node.attr["Tinput"])
            if "Tfilter" in quantized_node.attr:
                new_node.attr["Tfilter"].CopyFrom(quantized_node.attr["Tfilter"])
            if "strides" in quantized_node.attr:
                new_node.attr["strides"].CopyFrom(quantized_node.attr["strides"])
            if "padding" in quantized_node.attr:
                new_node.attr["padding"].CopyFrom(quantized_node.attr["padding"])
            if "alpha" in quantized_node.attr:
                new_node.attr["alpha"].CopyFrom(quantized_node.attr["alpha"])
            if "Tsummand" in quantized_node.attr:
                new_node.attr["Tsummand"].CopyFrom(quantized_node.attr["Tsummand"])
            if "data_format" in quantized_node.attr:
                new_node.attr["data_format"].CopyFrom(quantized_node.attr["data_format"])

            parent_node_name = Helper.node_name_from_input(quantized_node.input[0])
            if new_node.op in ("_FusedQuantizedDeconv2D", "_FusedQuantizedDeconv3D"):
                max_filter_node = self.graph_info[new_node.input[-3]].node
                min_filter_node = self.graph_info[new_node.input[-4]].node
            else:
                max_filter_node = self.graph_info[new_node.input[-1]].node
                min_filter_node = self.graph_info[new_node.input[-2]].node
            last_node = self.graph_info[new_node.input[0]].node
            new_node.input.append(requested_output_min_name)
            new_node.input.append(requested_output_max_name)

            if (
                last_node.op.find("Requantize") != -1
                or ((last_node.op.find("QuantizeV2") != -1 or last_node.op.find("QuantizedConv2D") != -1))
            ) and len(quantized_node.attr["fused_ops"].list.s) > 0:
                bias_node = self.graph_info[new_node.input[2]].node
                max_input_node = self.graph_info[last_node.input[-1]].node
                min_input_node = self.graph_info[last_node.input[-2]].node
                min_input = (min_input_node.attr["value"].tensor.float_val)[0]
                max_input = (max_input_node.attr["value"].tensor.float_val)[0]
                if "Depthwise" in quantized_node_op or requantize_node.op.find("PerChannel") != -1:
                    channel_size = max_filter_node.attr["value"].tensor.tensor_shape.dim[0].size
                    max_filter_tensor = tensor_util.MakeNdarray(min_filter_node.attr["value"].tensor)
                    min_filter_tensor = tensor_util.MakeNdarray(min_filter_node.attr["value"].tensor)
                else:
                    channel_size = 1
                    max_filter_tensor = []
                    min_filter_tensor = []
                    max_filter_tensor.append((max_filter_node.attr["value"].tensor.float_val)[0])
                    min_filter_tensor.append((min_filter_node.attr["value"].tensor.float_val)[0])
                bias_tensor = tensor_util.MakeNdarray(self.graph_info[new_node.input[2]].node.attr["value"].tensor)

                activation_range = 127.0 if new_node.attr["Tinput"].type == dtypes.qint8 else 255.0

                int32_bias = Helper.generate_int32_bias_for_conv(
                    bias_tensor,
                    channel_size,
                    max_input,
                    min_input,
                    max_filter_tensor,
                    min_filter_tensor,
                    activation_range,
                )

                bias_node.attr["dtype"].CopyFrom(
                    attr_value_pb2.AttrValue(type=float32_type if self.device == "gpu" else qint32_type)
                )

                bias_node.attr["value"].CopyFrom(
                    attr_value_pb2.AttrValue(
                        tensor=tensor_util.make_tensor_proto(
                            bias_tensor if self.device == "gpu" else int32_bias,
                            dtypes.float32 if self.device == "gpu" else dtypes.int32,
                            bias_tensor.shape,
                        )
                    )
                )

                bias_node.attr["value"].tensor.dtype = float32_type if self.device == "gpu" else qint32_type
                new_node.attr["Tbias"].CopyFrom(
                    attr_value_pb2.AttrValue(type=float32_type if self.device == "gpu" else qint32_type)
                )
            else:
                new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))
            # in tf 2.10, the "padding_list" attr name changes to explicit_paddings
            if "padding_list" in quantized_node.attr:
                if not self.new_api:
                    new_node.attr["padding_list"].CopyFrom(quantized_node.attr["padding_list"])
                elif quantized_node.attr["padding"].s == b"EXPLICIT":
                    new_node.attr["explicit_paddings"].CopyFrom(quantized_node.attr["padding_list"])
            elif "explicit_paddings" in quantized_node.attr:
                new_node.attr["explicit_paddings"].CopyFrom(quantized_node.attr["explicit_paddings"])
            if "dilations" in quantized_node.attr:
                new_node.attr["dilations"].CopyFrom(quantized_node.attr["dilations"])

            if self.new_api and new_node.op in (
                "_FusedQuantizedConv2D",
                "_FusedQuantizedDepthwiseConv2D",
                "_FusedQuantizedDeconv2D",
                "_FusedQuantizedDeconv3D",
            ):
                input_data_type = dtypes.qint8 if new_node.attr["Tinput"].type == dtypes.qint8 else dtypes.quint8
                if new_node.op in ("_FusedQuantizedDeconv2D", "_FusedQuantizedDeconv3D"):
                    Helper.set_attr_type_list(
                        new_node,
                        "Thost_inputs",
                        [
                            dtypes.int32.as_datatype_enum,
                            dtypes.qint8.as_datatype_enum,
                            input_data_type.as_datatype_enum,
                            (
                                dtypes.float32.as_datatype_enum
                                if new_node.attr["Tbias"].type == dtypes.float32
                                else dtypes.qint32.as_datatype_enum
                            ),
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                        ],
                    )
                else:
                    Helper.set_attr_type_list(
                        new_node,
                        "Thost_inputs",
                        [
                            input_data_type.as_datatype_enum,
                            dtypes.qint8.as_datatype_enum,
                            (
                                dtypes.float32.as_datatype_enum
                                if new_node.attr["Tbias"].type == dtypes.float32
                                else dtypes.qint32.as_datatype_enum
                            ),
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                        ],
                    )

                if quantized_node_op not in (
                    "_FusedQuantizedConv2D",
                    "_FusedQuantizedDepthwiseConv2D",
                    "_FusedQuantizedDeconv2D",
                    "_FusedQuantizedDeconv3D",
                ):
                    Helper.set_attr_type_list(new_node, "Thost_outputs", self.output_types)
                    new_node.attr["Tsummand"].CopyFrom(attr_value_pb2.AttrValue(type=self.output_types[0]))
                else:
                    if str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"_FusedHardSwish"]):
                        self.fused_ops = [b"BiasAdd", b"_FusedHardSwish", b"Requantize"]
                        Helper.set_attr_type_list(
                            new_node,
                            "Thost_outputs",
                            [
                                requantize_node.attr["out_type"].type,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                            ],
                        )
                        Helper.set_attr_dtype(
                            new_node, "out_type", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                        Helper.set_attr_dtype(
                            new_node, "Tsummand", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                    elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"_FusedSwish"]):
                        self.fused_ops = [b"BiasAdd", b"_FusedSwish", b"Requantize"]
                        Helper.set_attr_type_list(
                            new_node,
                            "Thost_outputs",
                            [
                                requantize_node.attr["out_type"].type,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                            ],
                        )
                        Helper.set_attr_dtype(
                            new_node, "out_type", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                        Helper.set_attr_dtype(
                            new_node, "Tsummand", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                    elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"Relu"]):
                        self.fused_ops = [b"BiasAdd", b"Relu", b"Requantize"]
                        Helper.set_attr_type_list(
                            new_node,
                            "Thost_outputs",
                            [
                                requantize_node.attr["out_type"].type,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                            ],
                        )
                        Helper.set_attr_dtype(
                            new_node, "out_type", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                        Helper.set_attr_dtype(
                            new_node, "Tsummand", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                    elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"LeakyRelu"]):
                        self.fused_ops = [b"BiasAdd", b"LeakyRelu", b"Requantize"]
                        Helper.set_attr_type_list(
                            new_node,
                            "Thost_outputs",
                            [
                                requantize_node.attr["out_type"].type,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                            ],
                        )
                        Helper.set_attr_dtype(
                            new_node, "out_type", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                        Helper.set_attr_dtype(
                            new_node, "Tsummand", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                    elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"Elu"]):
                        self.fused_ops = [b"BiasAdd", b"Elu", b"Requantize"]
                        Helper.set_attr_type_list(
                            new_node,
                            "Thost_outputs",
                            [
                                requantize_node.attr["out_type"].type,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                            ],
                        )
                        Helper.set_attr_dtype(
                            new_node, "out_type", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                        Helper.set_attr_dtype(
                            new_node, "Tsummand", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                    elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"Sigmoid"]):
                        self.fused_ops = [b"BiasAdd", b"Sigmoid", b"Requantize"]
                        Helper.set_attr_type_list(
                            new_node,
                            "Thost_outputs",
                            [
                                requantize_node.attr["out_type"].type,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                            ],
                        )
                        Helper.set_attr_dtype(
                            new_node, "out_type", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                        Helper.set_attr_dtype(
                            new_node, "Tsummand", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                    elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd"]):
                        self.fused_ops = [b"BiasAdd", b"Requantize"]
                        Helper.set_attr_type_list(
                            new_node,
                            "Thost_outputs",
                            [
                                requantize_node.attr["out_type"].type,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                            ],
                        )
                        Helper.set_attr_dtype(
                            new_node, "out_type", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                        if new_node.op not in ("_FusedQuantizedDeconv2D", "_FusedQuantizedDeconv3D"):
                            Helper.set_attr_dtype(
                                new_node, "Tsummand", dtype_map_dict[requantize_node.attr["out_type"].type]
                            )
                    elif len(quantized_node.attr["fused_ops"].list.s) == 0:
                        if new_node.op in ("_FusedQuantizedDeconv2D", "_FusedQuantizedDeconv3D"):
                            Helper.set_attr_type_list(
                                new_node,
                                "Thost_inputs",
                                [
                                    dtypes.int32.as_datatype_enum,
                                    dtypes.qint8.as_datatype_enum,
                                    input_data_type.as_datatype_enum,
                                    dtypes.float32.as_datatype_enum,
                                    dtypes.float32.as_datatype_enum,
                                    dtypes.float32.as_datatype_enum,
                                    dtypes.float32.as_datatype_enum,
                                    dtypes.float32.as_datatype_enum,
                                    dtypes.float32.as_datatype_enum,
                                ],
                            )
                        else:
                            Helper.set_attr_type_list(
                                new_node,
                                "Thost_inputs",
                                [
                                    input_data_type.as_datatype_enum,
                                    dtypes.qint8.as_datatype_enum,
                                    # dtypes.float32.as_datatype_enum if new_node.attr["Tbias"].type == dtypes.float32 \
                                    # else dtypes.qint32.as_datatype_enum,
                                    dtypes.float32.as_datatype_enum,
                                    dtypes.float32.as_datatype_enum,
                                    dtypes.float32.as_datatype_enum,
                                    dtypes.float32.as_datatype_enum,
                                    dtypes.float32.as_datatype_enum,
                                    dtypes.float32.as_datatype_enum,
                                ],
                            )
                        self.fused_ops = [b"Requantize"]
                        Helper.set_attr_type_list(
                            new_node,
                            "Thost_outputs",
                            [
                                requantize_node.attr["out_type"].type,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                            ],
                        )
                        Helper.set_attr_dtype(
                            new_node, "out_type", dtype_map_dict[requantize_node.attr["out_type"].type]
                        )
                        if new_node.op not in ("_FusedQuantizedDeconv2D", "_FusedQuantizedDeconv3D"):
                            Helper.set_attr_dtype(
                                new_node, "Tsummand", dtype_map_dict[requantize_node.attr["out_type"].type]
                            )
                Helper.set_attr_string_list(new_node, "fused_ops", self.fused_ops)

            if "_kernel" in quantized_node.attr:
                new_node.attr["_kernel"].CopyFrom(quantized_node.attr["_kernel"])

            if new_node.op in ("_FusedQuantizedConv3D"):
                input_data_type = dtypes.qint8 if new_node.attr["Tinput"].type == dtypes.qint8 else dtypes.quint8
                if len(quantized_node.attr["fused_ops"].list.s) == 0:
                    Helper.set_attr_string_list(new_node, "fused_ops", [b"Requantize"])
                    Helper.set_attr_type_list(
                        new_node,
                        "Thost_inputs",
                        [
                            input_data_type.as_datatype_enum,
                            dtypes.qint8.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                        ],
                    )
                elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd"]):
                    Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"Requantize"])
                    Helper.set_attr_type_list(
                        new_node,
                        "Thost_inputs",
                        [
                            input_data_type.as_datatype_enum,
                            dtypes.qint8.as_datatype_enum,
                            (
                                dtypes.float32.as_datatype_enum
                                if new_node.attr["Tbias"].type == dtypes.float32
                                else dtypes.qint32.as_datatype_enum
                            ),
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                        ],
                    )
                elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"Relu"]):
                    Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"Relu", b"Requantize"])
                    Helper.set_attr_type_list(
                        new_node,
                        "Thost_inputs",
                        [
                            input_data_type.as_datatype_enum,
                            dtypes.qint8.as_datatype_enum,
                            (
                                dtypes.float32.as_datatype_enum
                                if new_node.attr["Tbias"].type == dtypes.float32
                                else dtypes.qint32.as_datatype_enum
                            ),
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                        ],
                    )
                elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"LeakyRelu"]):
                    Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"LeakyRelu", b"Requantize"])
                    Helper.set_attr_type_list(
                        new_node,
                        "Thost_inputs",
                        [
                            input_data_type.as_datatype_enum,
                            dtypes.qint8.as_datatype_enum,
                            (
                                dtypes.float32.as_datatype_enum
                                if new_node.attr["Tbias"].type == dtypes.float32
                                else dtypes.qint32.as_datatype_enum
                            ),
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                        ],
                    )
                elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"Elu"]):
                    Helper.set_attr_string_list(new_node, "fused_ops", [b"BiasAdd", b"Elu", b"Requantize"])
                    Helper.set_attr_type_list(
                        new_node,
                        "Thost_inputs",
                        [
                            input_data_type.as_datatype_enum,
                            dtypes.qint8.as_datatype_enum,
                            (
                                dtypes.float32.as_datatype_enum
                                if new_node.attr["Tbias"].type == dtypes.float32
                                else dtypes.qint32.as_datatype_enum
                            ),
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                        ],
                    )

                Helper.set_attr_type_list(
                    new_node,
                    "Thost_outputs",
                    [
                        requantize_node.attr["out_type"].type,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ],
                )
                Helper.set_attr_dtype(new_node, "out_type", dtype_map_dict[requantize_node.attr["out_type"].type])
                Helper.set_attr_dtype(new_node, "Tsummand", dtype_map_dict[requantize_node.attr["out_type"].type])

            if (
                quantized_node.op == "QuantizedConv2D"
                or quantized_node.op == "QuantizedConv2DWithBias"
                or quantized_node.op == "QuantizedDepthwiseConv2D"
                or quantized_node.op == "QuantizedDepthwiseConv2DWithBias"
                or ("alpha" in quantized_node.attr and quantized_node.attr["alpha"].f > 0)
            ):
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=int8_type))
            elif (
                quantized_node.op == "QuantizedConv2DWithBiasAndRelu"
                or quantized_node.op == "QuantizedDepthwiseConv2DWithBiasAndRelu"
            ):
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))
            elif new_node.op not in (
                "_FusedQuantizedConv2D",
                "_FusedQuantizedDepthwiseConv2D",
                "_FusedQuantizedConv3D",
                "_FusedQuantizedDeconv2D",
                "_FusedQuantizedDeconv3D",
            ):
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))
            elif new_node.op in ("_FusedQuantizedDeconv2D", "_FusedQuantizedDeconv3D"):
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=int8_type))
            old_input_name = dequantize_node_name if dequantize_node_name else requantize_node_name
            self.graph_analyzer.replace_single_node(
                new_node,
                [parent_node_name],
                quantized_node_name,
                self.graph_info[old_input_name].outputs,
                old_input_name,
            )
            self.graph_analyzer.remove_node(quantized_node_name)

        target_nodes = self.graph_analyzer.query_fusion_pattern_nodes(self.sum_pattern)
        for i in target_nodes:
            quantized_node_name = i[0]
            quantized_node = self.graph_info[quantized_node_name].node
            if i[-1][0] in ("_FusedQuantizedDepthwiseConv2D", "_FusedQuantizedConv2D", "_FusedQuantizedConv3D"):
                if quantized_node.attr["fused_ops"].list.s not in self.fuse_sum_op_types:
                    continue
                # else:
                #   print(quantized_node.attr['fused_ops'].list.s)

            requantize_node_name = i[1]
            requantize_node = self.graph_info[requantize_node_name].node
            requested_output_min_name = requantize_node.input[3]
            requested_output_max_name = requantize_node.input[4]

            quantized_node_op = i[-1][0]

            new_node = node_def_pb2.NodeDef()

            if self.new_api:
                if i[-1][0] in ("QuantizedConv2DWithBiasSumAndRelu",):
                    new_node.op = "_FusedQuantizedConv2D"
                    self.fused_ops = [b"BiasAdd", b"Sum", b"Relu", b"Requantize"]
                elif i[-1][0] == "_FusedQuantizedConv2D":
                    new_node.op = "_FusedQuantizedConv2D"
                elif quantized_node_op == "_FusedQuantizedDepthwiseConv2D":
                    new_node.op = "_FusedQuantizedDepthwiseConv2D"
                elif i[-1][0] == "_FusedQuantizedConv3D":
                    new_node.op = "_FusedQuantizedConv3D"
            else:
                new_node.op = quantized_node_op + "AndRequantize"

            new_node.name = requantize_node_name

            for _, value in enumerate(quantized_node.input[:-1]):
                new_node.input.append(value)

            new_node.attr["Tinput"].CopyFrom(quantized_node.attr["Tinput"])
            new_node.attr["Tfilter"].CopyFrom(quantized_node.attr["Tfilter"])
            new_node.attr["strides"].CopyFrom(quantized_node.attr["strides"])
            new_node.attr["padding"].CopyFrom(quantized_node.attr["padding"])
            # new_node.attr["Tsummand"].CopyFrom(quantized_node.attr['Tsummand'])

            new_node.input.append(requested_output_min_name)
            new_node.input.append(requested_output_max_name)
            deq_node = self.graph_info[Helper.node_name_from_input(quantized_node.input[-1])].node
            if deq_node.op != "Dequantize" or deq_node.op.find("Quantize") != -1:
                continue

            if deq_node.op == "Dequantize":
                original_summand_node = self.graph_info[Helper.node_name_from_input(deq_node.input[0])].node
            else:
                original_summand_node = deq_node
            summand_op_type = uint8_type if dtypes.as_dtype(deq_node.attr["T"].type) == uint8_type else int8_type

            for j in range(3):
                new_node.input.append(original_summand_node.name + ":{}".format(j))
            # in tf 2.10, the "padding_list" attr name changes to explicit_paddings
            if "padding_list" in quantized_node.attr:
                if not self.new_api:
                    new_node.attr["padding_list"].CopyFrom(quantized_node.attr["padding_list"])
                elif quantized_node.attr["padding"].s == b"EXPLICIT":
                    new_node.attr["explicit_paddings"].CopyFrom(quantized_node.attr["padding_list"])
            elif "explicit_paddings" in quantized_node.attr:
                new_node.attr["explicit_paddings"].CopyFrom(quantized_node.attr["explicit_paddings"])

            if "dilations" in quantized_node.attr:
                new_node.attr["dilations"].CopyFrom(quantized_node.attr["dilations"])

            if "alpha" in quantized_node.attr and quantized_node.attr["alpha"].f > 0:
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=int8_type))
            else:
                new_node.attr["out_type"].CopyFrom(attr_value_pb2.AttrValue(type=uint8_type))

            new_node.attr["Tbias"].CopyFrom(attr_value_pb2.AttrValue(type=float32_type))
            new_node.attr["Tsummand"].CopyFrom(attr_value_pb2.AttrValue(type=summand_op_type))

            if new_node.op in ("_FusedQuantizedConv2D", "_FusedQuantizedDepthwiseConv2D", "_FusedQuantizedConv3D"):
                original_input = list(new_node.input)
                new_input = []
                new_input.extend(original_input[:3])
                new_input.append(original_input[-3])
                new_input.extend(original_input[3:7])
                new_input.extend(original_input[-2:])
                new_input.extend(original_input[-5:-3])
                new_node.ClearField("input")
                new_node.input.extend(new_input)
                input_data_type = dtypes.qint8 if new_node.attr["Tinput"].type == dtypes.qint8 else dtypes.quint8
                Helper.set_attr_type_list(
                    new_node,
                    "Thost_inputs",
                    [
                        input_data_type.as_datatype_enum,
                        dtypes.qint8.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        (
                            dtypes.quint8.as_datatype_enum
                            if summand_op_type != int8_type
                            else dtypes.qint8.as_datatype_enum
                        ),
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ],
                )
                Helper.set_attr_dtype(
                    new_node, "Tsummand", dtypes.quint8 if summand_op_type != int8_type else dtypes.qint8
                )
                if str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"Sum", b"Relu"]):
                    self.fused_ops = [b"BiasAdd", b"Sum", b"Relu", b"Requantize"]
                elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"Sum", b"LeakyRelu"]):
                    self.fused_ops = [b"BiasAdd", b"Sum", b"LeakyRelu", b"Requantize"]
                elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"LeakyRelu", b"Sum"]):
                    self.fused_ops = [b"BiasAdd", b"LeakyRelu", b"Sum", b"Requantize"]
                elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"Relu", b"Sum"]):
                    self.fused_ops = [b"BiasAdd", b"Relu", b"Sum", b"Requantize"]
                    # Current fusion requires summand has same dtype as output if output is qint8
                    Helper.set_attr_dtype(new_node, "Tsummand", dtype_map_dict[requantize_node.attr["out_type"].type])
                elif str(quantized_node.attr["fused_ops"].list.s) == str([b"BiasAdd", b"Sum"]):
                    self.fused_ops = [b"BiasAdd", b"Sum", b"Requantize"]
                    # Current fusion requires summand has same dtype as output if output is qint8
                    Helper.set_attr_dtype(new_node, "Tsummand", dtype_map_dict[requantize_node.attr["out_type"].type])
                Helper.set_attr_type_list(
                    new_node,
                    "Thost_outputs",
                    [
                        requantize_node.attr["out_type"].type,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ],
                )
                Helper.set_attr_dtype(new_node, "out_type", dtype_map_dict[requantize_node.attr["out_type"].type])
                Helper.set_attr_string_list(new_node, "fused_ops", self.fused_ops)

            if not self.new_api:
                if quantized_node_op == "QuantizedConv2DWithBiasReluAndSum":
                    new_node.op = "QuantizedConv2DWithBiasReluAndSumAndRequantize"
                    if "alpha" in quantized_node.attr:
                        new_node.attr["alpha"].CopyFrom(quantized_node.attr["alpha"])

                elif summand_op_type == int8_type:
                    new_node.op = "QuantizedConv2DWithBiasSignedSumAndReluAndRequantize"

            self.graph_analyzer.replace_single_node(
                new_node,
                [quantized_node.input[0], original_summand_node.name],
                quantized_node.name,
                self.graph_info[requantize_node_name].outputs,
                requantize_node_name,
            )
            self.graph_analyzer.remove_node(quantized_node_name)

            if deq_node.op == "Dequantize":
                self.graph_analyzer.remove_node_with_single_input_output(deq_node.name)

        return self.graph_analyzer.dump_graph()



#
#  -*- coding: utf-8 -*-
#
#  Copyright (c) 2021 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""Bias correction graph transform."""

from __future__ import absolute_import, division, print_function

import numpy as np
from tensorflow.core.framework import attr_value_pb2
from tensorflow.python.framework import dtypes, tensor_util

from .graph_transform_base import GraphTransformBase


class BiasCorrection(GraphTransformBase):
    """This class implements the bias correction graph transform.

    Will correct the weight and scale for *Conv2D* op
    weight_empirical:
    our task is to correct int8 weight distribution close to fp32 weight
    r*(W_int8 + u) -> W_fp32, r is variance ratio between fp32 and int8
    u is the difference between fp32 and int8 channel wise, it's equal to minimize:
    round(scale_c * (W_fp32 + shift))/scale - r*(round(scale * W_fp32) + scale*u)/scale
    notice we can only change the first round: round(scale_c * (W_fp32 + shift))
    an empirical solution is to make:
    scale_c = r * scale and shift = u
    with this we don't change the min/max value, and correct the weight
    """

    def __init__(self, input_graph, fp32_graph, method="weight_empirical", new_api=False):
        """Initialization."""
        # only support weight_empirical now
        self.bias_correct_map = {"weight_empirical": self._weight_empirical}
        assert method in self.bias_correct_map, "only support weight empirical correction method"

        super(BiasCorrection, self).__init__(input_graph)
        self.fp32_graph = fp32_graph
        self.input_graph = input_graph
        self.method = method
        self.fp32_node_mapping = {}
        self.parse_input_pb()
        self.new_api = new_api

    def _weight_empirical(self):
        """Weight empirical correction method."""
        for node in self.fp32_graph.node:
            if node.name not in self.fp32_node_mapping:
                self.fp32_node_mapping[node.name] = node

        for node_name in self.node_mapping:
            node = self.node_mapping[node_name]
            node_op = node.op
            if self.new_api:
                if "_FusedQuantizedConv2D" not in node_op:
                    continue
            else:
                if "QuantizedConv2D" not in node_op:
                    continue

            int8_filter = self.node_mapping[self.get_node_name_from_input(node.input[1])]

            int8_value = tensor_util.MakeNdarray(int8_filter.attr["value"].tensor)
            tr_int8_value = int8_value.transpose([3, 0, 1, 2])

            fp32_filter_name = self.get_node_name_from_input(node.input[1]).split("_qint8_const")[0]
            fp32_filter = self.fp32_node_mapping[fp32_filter_name]

            fp32_value = tensor_util.MakeNdarray(fp32_filter.attr["value"].tensor)
            tr_fp32_value = fp32_value.transpose([3, 0, 1, 2])

            # if bias fused, then offset to min/max filter should be 5
            if self.new_api:
                offset = 5
            else:
                offset = 5 if "Bias" in node_op else 4
            min_filter_node = self.node_mapping[node.input[offset]]
            max_filter_node = self.node_mapping[node.input[offset + 1]]

            channel_size = (
                1
                if not min_filter_node.attr["value"].tensor.tensor_shape.dim
                else min_filter_node.attr["value"].tensor.tensor_shape.dim[0].size
            )

            if channel_size == 1:
                max_filter_tensor = []
                min_filter_tensor = []
                max_filter_tensor.append((max_filter_node.attr["value"].tensor.float_val)[0])
                min_filter_tensor.append((min_filter_node.attr["value"].tensor.float_val)[0])
            else:
                max_filter_tensor = tensor_util.MakeNdarray(max_filter_node.attr["value"].tensor)
                min_filter_tensor = tensor_util.MakeNdarray(min_filter_node.attr["value"].tensor)

            tr_quantized_fp32_value = np.zeros_like(tr_fp32_value)
            tr_corrected_int8_value = np.zeros_like(tr_int8_value)

            for i in range(channel_size):
                scale = max(abs(max_filter_tensor[i]), abs(min_filter_tensor[i])) / 127
                tr_quantized_fp32_value[i] = tr_int8_value[i].astype(np.float64) * scale
                delta_mean = np.mean((tr_fp32_value[i] - tr_quantized_fp32_value[i]).flatten())
                var_ratio = (
                    np.std(tr_fp32_value[i].flatten()) / np.std(tr_quantized_fp32_value[i].flatten())
                    if np.std(tr_quantized_fp32_value[i].flatten()) != 0
                    else 1
                )
                tr_corrected_int8_value[i] = (var_ratio / scale) * (tr_fp32_value[i] + delta_mean)

            correct_int8_value = tr_int8_value.transpose([1, 2, 3, 0])
            assert (
                int8_value.shape == correct_int8_value.shape
            ), "correct filter shape should equal with origin filter shape"
            bias = int8_value.astype(np.float32) - correct_int8_value.astype(np.float32)
            if np.sum(bias) != 0:
                int8_filter.attr["value"].CopyFrom(
                    attr_value_pb2.AttrValue(
                        tensor=tensor_util.make_tensor_proto(correct_int8_value, dtypes.qint8, int8_value.shape)
                    )
                )
        return self.input_graph

    def do_transformation(self):
        """Apply bias correction graph transform."""
        return self.bias_correct_map[self.method]()
